# -*- coding: utf-8 -*-
#
#  Copyright 2019-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import Optional, Tuple, Union, List, Set
from weakref import ref
from ..exceptions import IsConnectedBond, IsNotConnectedBond


class Bond:
    __slots__ = ('__order', '__graph', '__n', '__m')

    def __init__(self, order: int):
        if not isinstance(order, int):
            raise TypeError('invalid order value')
        elif order not in (1, 4, 2, 3, 8):
            raise ValueError('order should be from [1, 2, 3, 4, 8]')
        self.__order = order

    def __eq__(self, other):
        if isinstance(other, Bond):
            return self.__order == other.order
        elif isinstance(other, int):
            return self.__order == other
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__order})'

    def __int__(self):
        """
        Bond order.
        """
        return self.__order

    def __hash__(self):
        """
        Bond order. Used in Morgan atoms ordering.
        """
        return self.__order

    def __getstate__(self):
        return {'order': self.__order}

    def __setstate__(self, state):
        self.__order = state['order']

    @property
    def order(self) -> int:
        return self.__order

    @property
    def in_ring(self) -> bool:
        try:
            return self.__graph().is_ring_bond(self.__n, self.__m)
        except AttributeError:
            raise IsNotConnectedBond

    def copy(self) -> 'Bond':
        copy = object.__new__(self.__class__)
        copy._Bond__order = self.__order
        return copy

    @classmethod
    def from_bond(cls, bond):
        if isinstance(bond, cls):
            copy = object.__new__(cls)
            copy._Bond__order = bond.order
            return copy
        raise TypeError('Bond expected')

    def _attach_graph(self, graph, n, m):
        try:
            self.__graph
        except AttributeError:
            self.__graph = ref(graph)
            self.__n = n
            self.__m = m
        else:
            raise IsConnectedBond

    def _change_map(self, n, m):
        try:
            self.__graph
        except AttributeError:
            raise IsNotConnectedBond
        else:
            self.__n = n
            self.__m = m


class DynamicBond:
    __slots__ = ('__order', '__p_order')

    def __init__(self, order=None, p_order=None):
        if order is None:
            if not isinstance(p_order, int):
                raise TypeError('p_order should be int type')
        elif not isinstance(order, int):
            raise TypeError('order should be int type or None')
        elif p_order is not None and not isinstance(p_order, int):
            raise TypeError('p_order should be int type or None')

        if order not in (1, 4, 2, 3, None, 8) or p_order not in (1, 4, 2, 3, None, 8):
            raise ValueError('order or p_order should be from [1, 2, 3, 4, 8]')

        self.__order = order
        self.__p_order = p_order

    def __eq__(self, other):
        if isinstance(other, DynamicBond):
            return self.__order == other.order and self.__p_order == other.p_order
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__order}, {self.__p_order})'

    def __int__(self):
        """
        Hash of bond orders.
        """
        return hash(self)

    def __hash__(self):
        """
        Hash of bond orders.
        """
        return hash((self.__order or 0, self.__p_order or 0))

    @property
    def is_dynamic(self) -> bool:
        """
        Bond has dynamic features
        """
        return self.__order != self.__p_order

    @property
    def order(self) -> Optional[int]:
        return self.__order

    @property
    def p_order(self) -> Optional[int]:
        return self.__p_order

    def copy(self) -> 'DynamicBond':
        copy = object.__new__(self.__class__)
        copy._DynamicBond__order = self.__order
        copy._DynamicBond__p_order = self.__p_order
        return copy

    @classmethod
    def from_bond(cls, bond):
        if isinstance(bond, Bond):
            copy = object.__new__(cls)
            copy._DynamicBond__order = copy._DynamicBond__p_order = bond.order
            return copy
        elif isinstance(bond, cls):
            copy = object.__new__(cls)
            copy._DynamicBond__order = bond.order
            copy._DynamicBond__p_order = bond.p_order
            return copy
        raise TypeError('DynamicBond expected')


class QueryBond:
    __slots__ = ('__order', '__in_ring')

    def __init__(self, order: Union[int, List[int], Set[int], Tuple[int, ...]], in_ring: Optional[bool] = None):
        if isinstance(order, (list, tuple, set)):
            if not all(isinstance(x, int) for x in order):
                raise TypeError('invalid order value')
            if any(x not in (1, 4, 2, 3, 8) for x in order):
                raise ValueError('order should be from [1, 2, 3, 4, 8]')
            order = tuple(sorted(set(order)))
        elif isinstance(order, int):
            if order not in (1, 4, 2, 3, 8):
                raise ValueError('order should be from [1, 2, 3, 4, 8]')
            order = (order,)
        else:
            raise TypeError('invalid order value')
        if in_ring is not None and not isinstance(in_ring, bool):
            raise TypeError('in_ring mark should be boolean or None')
        self.__order = order
        self.__in_ring = in_ring

    def __eq__(self, other):
        if isinstance(other, Bond):
            if self.__in_ring is not None:
                if self.__in_ring != other.in_ring:
                    return False
            return other.order in self.__order
        elif isinstance(other, QueryBond):
            return self.__order == other.order and self.__in_ring == other.in_ring
        elif isinstance(other, int):
            return other in self.__order
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__order}, {self.__in_ring})'

    def __int__(self):
        """
        Simple bond order or hash of sorted tuple of orders.
        """
        if len(self.__order) == 1:
            return self.__order[0]
        return hash(self.__order)

    def __hash__(self):
        """
        Hash of orders and cycle mark. Used in Morgan atoms ordering.
        """
        return hash((self.__order, self.__in_ring))

    @property
    def order(self) -> Tuple[int, ...]:
        return self.__order

    @property
    def in_ring(self) -> Optional[bool]:
        return self.__in_ring

    def copy(self) -> 'QueryBond':
        copy = object.__new__(self.__class__)
        copy._QueryBond__order = self.__order
        copy._QueryBond__in_ring = self.__in_ring
        return copy

    @classmethod
    def from_bond(cls, bond):
        if isinstance(bond, Bond):
            copy = object.__new__(cls)
            copy._QueryBond__order = (bond.order,)
            copy._QueryBond__in_ring = None
            return copy
        elif isinstance(bond, cls):
            copy = object.__new__(cls)
            copy._QueryBond__order = bond.order
            copy._QueryBond__in_ring = bond.in_ring
            return copy
        raise TypeError('QueryBond or Bond expected')


__all__ = ['Bond', 'DynamicBond', 'QueryBond']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Alexander Sizov <murkyrussian@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from chython import MoleculeContainer
from typing import Tuple, List
from ..algorithms.depict import _render_config, _graph_svg


Tree = Tuple[MoleculeContainer, List['Tree']]


def retro_depict(tree: Tree, *, y_gap=3., x_gap=5.) -> str:
    """
    Depict retrosynthetic tree.

    :param tree: Graph of molecules with recursive structure. Each node is tuple of molecule and list of child nodes.
        Child nodes list can be empty.
    :param y_gap: vertical gap between molecules.
    :param x_gap: horizontal gap between molecules.
    """
    font_size = _render_config['font_size']
    font125 = 1.25 * font_size

    arrows = []
    columns = [[tree[0]]]
    current_layer = [(0, x) for x in tree[1]]
    while current_layer:
        next_layer = []
        column = []
        columns.append(column)
        for i, (j, (m, ms)) in enumerate(current_layer):
            column.append(m)
            arrows.append(j)  # i-th molecule in current column connected to j-th molecule in previous.
            next_layer.extend((i, x) for x in ms)
        current_layer = next_layer

    x_shift = 0.
    c_max_x = 0.
    c_max_y = 0.
    arrows = iter(arrows)
    render = []
    last_layer = []
    arrows_coords = []
    for column in columns:
        current_layer = []
        heights = [max(y for _, y in m._plane.values()) - min(y for _, y in m._plane.values()) for m in column]
        y_shift = sum(heights) + y_gap * (len(heights) - 1)  # column height with gaps
        if y_shift > c_max_y:
            c_max_y = y_shift
        y_shift /= 2.  # center align

        for m, h in zip(column, heights):
            plane = m._plane.copy()  # backup
            mx = m._fix_plane_min(x_shift, -y_shift)
            if mx > c_max_x:
                c_max_x = mx

            current_layer.append((mx + 1., y_shift - h / 2.))
            if x_shift:  # except first column
                arrows_coords.append((*last_layer[next(arrows)], x_shift - 1., y_shift - h / 2.))
                y_shift -= h + y_gap

            render.append(m.depict(embedding=True)[:5])
            m._plane = plane  # restore

        x_shift = c_max_x + x_gap  # between columns gap
        last_layer = current_layer

    width = c_max_x + 3.0 * font_size
    height = c_max_y + 2.5 * font_size
    box_y = height / 2.
    svg = [f'<svg width="{width:.2f}cm" height="{height:.2f}cm" '
           f'viewBox="{-font125:.2f} {-box_y:.2f} {width:.2f} {height:.2f}" '
           'xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" version="1.1">',
           '  <defs>\n    <marker id="arrow" markerWidth="10" markerHeight="10" '
           'refX="0" refY="3" orient="auto">\n      <path d="M0,0 L0,6 L9,3"/>\n    </marker>\n  </defs>']
    for atoms, bonds, define, masks, uid in render:
        svg.extend(_graph_svg(atoms, bonds, define, masks, uid, -font125, -box_y, width, height))

    svg.append('  <g fill="none" stroke="black" stroke-width=".04" marker-end="url(#arrow)">')
    for x1, y1, x2, y2 in arrows_coords:
        svg.append(f'    <line x1="{x1:.2f}" y1="{y1:.2f}" x2="{x2:.2f}" y2="{y2:.2f}"/>')
    svg.append('  </g>')
    svg.append('</svg>')
    return '\n'.join(svg)


class RetroDepict:
    """
    Grid depict for Jupyter notebooks.
    """
    def __init__(self, tree: Tree):
        """
        :param tree: Graph of molecules with recursive structure.
            Each node is tuple of molecule and list of child nodes.
            Child nodes list can be empty.
        """
        self.tree = tree

    def _repr_svg_(self):
        return retro_depict(self.tree)


__all__ = ['retro_depict', 'RetroDepict']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from fileinput import FileInput
from io import StringIO, TextIOWrapper
from itertools import chain
from pathlib import Path
from re import compile, findall, search, split
from typing import List, Union, Optional
from .parser import parser
from .tokenize import smiles_tokenize
from .._mdl import Parser
from ...containers import MoleculeContainer, ReactionContainer
from ...exceptions import IsChiral, NotChiral, ValenceError, ParseError


delimiter = compile(r'[=:]')
cx_fragments = compile(r'f:(?:[0-9]+(?:\.[0-9]+)+)(?:,(?:[0-9]+(?:\.[0-9]+)+))*')
cx_radicals = compile(r'\^[1-7]:[0-9]+(?:,[0-9]+)*')

implicit_mismatch_key = 'implicit_mismatch'


class SMILESRead(Parser):
    """SMILES separated per lines files reader. Works similar to opened file object. Support `with` context manager.
    On initialization accept opened in text mode file, string path to file,
    pathlib.Path object or another buffered reader object.

    Line should be start with SMILES string and optionally continues with space/tab separated list of
    `key:value` [or `key=value`] data if `header=None`. For example::

        C=C>>CC id:123 key=value

    if `header=True` then first line of file should be space/tab separated list of keys including smiles column key.
    For example::

        ignored_smi_key key1 key2
        CCN 1 2

    Also possible to pass list of keys (without smiles_pseudo_key) for mapping space/tab separated list
    of SMILES and values: `header=['key1', 'key2'] # order depended`.

    For reactions . [dot] in bonds should be used only for molecules separation.
    """
    def __init__(self, file, header=None, ignore_stereo=False, keep_implicit=False, ignore_carbon_radicals=False,
                 **kwargs):
        """
        :param ignore: Skip some checks of data or try to fix some errors.
        :param remap: Remap atom numbers started from one.
        :param store_log: Store parser log if exists messages to `.meta` by key `ParserLog`.
        :param ignore_stereo: Ignore stereo data.
        :param keep_implicit: keep given in smiles implicit hydrogen count, otherwise ignore on valence error.
        :param ignore_bad_isotopes: reset invalid isotope mark to non-isotopic.
        :param ignore_carbon_radicals: fill carbon radicals with hydrogen (X[C](X)X case).
        """
        if isinstance(file, str):
            self._file = open(file)
            self.__is_buffer = False
        elif isinstance(file, Path):
            self._file = file.open()
            self.__is_buffer = False
        elif isinstance(file, (TextIOWrapper, StringIO, FileInput)):
            self._file = file
            self.__is_buffer = True
        else:
            raise TypeError('invalid file. TextIOWrapper, StringIO subclasses possible')
        super().__init__(**kwargs)
        self.__file = iter(self._file.readline, '')

        if header is True:
            self.__header = next(self.__file).split()[1:]
        elif header:
            if not isinstance(header, (list, tuple)) or not all(isinstance(x, str) for x in header):
                raise TypeError('expected list (tuple) of strings')
            self.__header = header
        else:
            self.__header = None

        self.__ignore_stereo = ignore_stereo
        self.__keep_implicit = keep_implicit
        self.__ignore_carbon_radicals = ignore_carbon_radicals
        self._data = self.__data()

    def __data(self):
        file = self._file
        parse = self.parse
        try:
            seekable = file.seekable()
        except AttributeError:
            seekable = False
        pos = file.tell() if seekable else None
        for n, line in enumerate(self.__file):
            try:
                x = parse(line)
            except ValueError:  # yield exception, not raise
                yield ParseError(n, pos, self._format_log(), line)
            else:
                yield x
            if seekable:
                pos = file.tell()

    @classmethod
    def create_parser(cls, header=None, ignore_stereo=False, keep_implicit=False, ignore_carbon_radicals=False,
                      **kwargs):
        """
        Create SMILES parser function configured same as SMILESRead object.
        """
        obj = object.__new__(cls)
        obj._SMILESRead__header = header
        obj._SMILESRead__ignore_stereo = ignore_stereo
        obj._SMILESRead__keep_implicit = keep_implicit
        obj._SMILESRead__ignore_carbon_radicals = ignore_carbon_radicals
        super(SMILESRead, obj).__init__(**kwargs)
        return obj.parse

    def close(self, force=False):
        """
        Close opened file.

        :param force: Force closing of externally opened file or buffer.
        """
        if not self.__is_buffer or force:
            self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def read(self) -> List[Union[MoleculeContainer, ReactionContainer]]:
        """
        Parse whole file.

        :return: List of parsed molecules or reactions.
        """
        return list(iter(self))

    def __iter__(self):
        return (x for x in self._data if not isinstance(x, ParseError))

    def __next__(self):
        return next(iter(self))

    def parse(self, smiles: str) -> Union[MoleculeContainer, ReactionContainer]:
        """
        SMILES string parser. String should start with SMILES and
        optionally continues with space/tab separated list of key:value [or key=value] data.
        """
        if not smiles:
            raise ValueError('Empty string')

        contract: Optional[List[List[int]]]  # typing

        smi, *data = smiles.split()
        if data and data[0].startswith('|') and data[0].endswith('|'):
            fr = search(cx_fragments, data[0])
            if fr is not None:
                contract = [sorted(int(x) for x in x.split('.')) for x in fr.group()[2:].split(',')]
                if len({x for x in contract for x in x}) < len([x for x in contract for x in x]):
                    self._info(f'collisions in cxsmiles fragments description: {data[0]}')
                    contract = None

                radicals = [int(x) for x in findall(cx_radicals, data[0]) for x in x[3:].split(',')]
                if radicals and len(set(radicals)) != len(radicals):
                    self._info(f'collisions in cxsmiles radicals description: {data[0]}')
                    radicals = []
                data.pop(0)
            else:
                radicals = [int(x) for x in findall(cx_radicals, data[0]) for x in x[3:].split(',')]
                if radicals:
                    if len(set(radicals)) != len(radicals):
                        self._info(f'collisions in cxsmiles radicals description: {data[0]}')
                        radicals = []
                    data.pop(0)
                contract = None
        else:
            radicals = []
            contract = None

        if self.__header is None:
            meta = {}
            for x in data:
                try:
                    k, v = split(delimiter, x, 1)
                    meta[k] = v
                except ValueError:
                    self._info(f'invalid metadata entry: {x}')
        else:
            meta = dict(zip(self.__header, data))

        if '>' in smi:
            record = {'reactants': [], 'reagents': [], 'products': [], 'meta': meta, 'title': ''}
            try:
                reactants, reagents, products = smi.split('>')
            except ValueError as e:
                raise ValueError('invalid reaction smiles') from e

            mol_count = 0
            for k, d in zip(('reactants', 'products', 'reagents'), (reactants, products, reagents)):
                if not d:
                    continue
                for x in d.split('.'):
                    if not x:
                        if self._ignore:
                            self._info('two dots in line ignored')
                        else:
                            raise ValueError('invalid reaction smiles. two dots in line')
                    else:
                        record[k].append(x)
                        mol_count += 1

            if contract:
                if max(x for x in contract for x in x) >= mol_count:
                    self._info(f'skipped invalid contract data: {contract}')
                lr = len(record['reactants'])
                lp = len(record['products'])
                reactants = set(range(lr))
                reagents = set(range(lr, mol_count - lp))
                products = set(range(mol_count - lp, mol_count))
                new_molecules: List[Optional[str]] = [None] * mol_count
                for c in contract:
                    if reactants.issuperset(c):
                        new_molecules[c[0]] = '.'.join(record['reactants'][x] for x in c)
                        reactants.difference_update(c)
                    elif products.issuperset(c):
                        new_molecules[c[0]] = '.'.join(record['products'][x - mol_count] for x in c)
                        products.difference_update(c)
                    elif reagents.issuperset(c):
                        new_molecules[c[0]] = '.'.join(record['reagents'][x - lr] for x in c)
                        reagents.difference_update(c)
                    else:
                        self._info(f'impossible to contract different parts of reaction: {contract}')
                for x in reactants:
                    new_molecules[x] = record['reactants'][x]
                for x in products:
                    new_molecules[x] = record['products'][x - mol_count]
                for x in reagents:
                    new_molecules[x] = record['reagents'][x - lr]

                record['reactants'] = [x for x in new_molecules[:lr] if x is not None]
                record['products'] = [x for x in new_molecules[-lp:] if x is not None]
                record['reagents'] = [x for x in new_molecules[lr: -lp] if x is not None]

            for k in ('reactants', 'products', 'reagents'):
                tmp = []
                for x in record[k]:
                    r, log = parser(smiles_tokenize(x), not self._ignore)
                    tmp.append(r)
                    for l in log:
                        self._info(l)
                record[k] = tmp

            if radicals:
                atom_map = dict(enumerate(a for m in chain(record['reactants'], record['reagents'], record['products'])
                                          for a in m['atoms']))
                for x in radicals:
                    atom_map[x]['is_radical'] = True

            return self._convert_reaction(record)
        else:
            record, log = parser(smiles_tokenize(smi), not self._ignore)
            for l in log:
                self._info(l)
            record['meta'].update(meta)
            for x in radicals:
                record['atoms'][x]['is_radical'] = True
            return self._convert_molecule(record)

    def _create_molecule(self, data, mapping):
        mol = super()._create_molecule(data, mapping)
        atoms = mol._atoms
        bonds = mol._bonds
        charges = mol._charges
        hydrogens = mol._hydrogens
        radicals = mol._radicals
        calc_implicit = mol._calc_implicit
        hyb = mol.hybridization
        keep_implicit = self.__keep_implicit
        radicalized = []

        for n, a in enumerate(data['atoms']):
            h = a['hydrogen']
            if h is None:  # simple atom token
                continue
            # bracket token should always contain implicit hydrogens count.
            n = mapping[n]
            if keep_implicit:  # override any calculated hydrogens count.
                hydrogens[n] = h
            elif (hc := hydrogens[n]) is None:  # atom has invalid valence for now.
                if hyb(n) == 4:  # this is aromatic rings. just store given H count.
                    hydrogens[n] = h
                    if not h and not charges[n] and atoms[n].atomic_number in (5, 6, 7, 15) and \
                            sum(b.order != 8 for b in bonds[n].values()) == 2:
                        # c[c]c - aromatic B,C,N,P radical
                        radicals[n] = True
                        radicalized.append(n)
                elif not radicals[n]:  # CXSMILES radical not set.
                    # SMILES doesn't code radicals. so, let's try to guess.
                    radicals[n] = True
                    calc_implicit(n)
                    if hydrogens[n] != h:  # radical state also has errors.
                        if self._ignore:
                            hydrogens[n] = None  # reset hydrogens
                            radicals[n] = False  # reset radical state
                            if implicit_mismatch_key in mol.meta:
                                mol.meta[implicit_mismatch_key][n] = h
                            else:
                                mol.meta[implicit_mismatch_key] = {n: h}
                            self._info(f'implicit hydrogen count ({h}) mismatch with calculated ({hc}) on atom {n}.')
                        else:
                            raise ValueError(f'implicit hydrogen count ({h}) mismatch with '
                                             f'calculated ({hc}) on atom {n}.')
                    else:
                        radicalized.append(n)
            elif hc != h and not radicals[n]:  # H count mismatch. try radical state of atom.
                radicals[n] = True
                calc_implicit(n)
                if hydrogens[n] != h:  # radical state also has errors.
                    if self._ignore:
                        hydrogens[n] = hc  # reset hydrogens to previous valid state
                        radicals[n] = False  # reset radical state
                        if implicit_mismatch_key in mol.meta:
                            mol.meta[implicit_mismatch_key][n] = h
                        else:
                            mol.meta[implicit_mismatch_key] = {n: h}
                        self._info(f'implicit hydrogen count ({h}) mismatch with calculated ({hc}) on atom {n}.')
                    else:
                        raise ValueError(f'implicit hydrogen count ({h}) mismatch with '
                                         f'calculated ({hc}) on atom {n}.')
                else:
                    radicalized.append(n)

        if self.__ignore_carbon_radicals:
            for n in radicalized:
                if atoms[n].atomic_number == 6:
                    radicals[n] = False
                    if (h := 4 - sum(b.order for b in bonds[n].values() if b.order != 8)) >= 0:
                        hydrogens[n] = h
                        self._info(f'Carbon radical replaced with ({h}) implicit hydrogens')

        if self.__ignore_stereo:
            return mol
        stereo_atoms = [(n, s) for n, a in enumerate(data['atoms']) if (s := a['stereo']) is not None]
        if not stereo_atoms and not data['stereo_bonds']:
            return mol

        st = mol._stereo_tetrahedrons
        sa = mol._stereo_allenes
        sat = mol._stereo_allenes_terminals
        ctt = mol._stereo_cis_trans_terminals

        order = {mapping[n]: [mapping[m] for m in ms] for n, ms in data['order'].items()}

        stereo = []
        for i, s in stereo_atoms:
            n = mapping[i]
            if not i and hydrogens[n]:  # first atom in smiles has reversed chiral mark
                s = not s

            if n in st:
                stereo.append((mol.add_atom_stereo, n, order[n], s))
            elif n in sa:
                t1, t2 = sat[n]
                env = sa[n]
                n1 = next(x for x in order[t1] if x in env)
                n2 = next(x for x in order[t2] if x in env)
                stereo.append((mol.add_atom_stereo, n, (n1, n2), s))

        stereo_bonds = {mapping[n]: {mapping[m]: s for m, s in ms.items()}
                        for n, ms in data['stereo_bonds'].items()}
        seen = set()
        for n, ns in stereo_bonds.items():
            if n in seen:
                continue
            if n in ctt:
                nm = ctt[n]
                m = nm[1] if nm[0] == n else nm[0]
                if m in stereo_bonds:
                    seen.add(m)
                    n2, s2 = stereo_bonds[m].popitem()
                    n1, s1 = ns.popitem()
                    stereo.append((mol.add_cis_trans_stereo, n, m, n1, n2, s1 == s2))

        while stereo:
            fail_stereo = []
            old_stereo = len(stereo)
            for f, *args in stereo:
                try:
                    f(*args, clean_cache=False)
                except NotChiral:
                    fail_stereo.append((f, *args))
                except IsChiral:
                    pass
                except ValenceError:
                    self._info('structure has errors, stereo data skipped')
                    mol.flush_cache()
                    break
            else:
                stereo = fail_stereo
                if len(stereo) == old_stereo:
                    break
                mol.flush_stereo_cache()
                continue
            break
        return mol


__all__ = ['SMILESRead']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from itertools import product
from typing import Union, List
from .retro import Tree
from ..containers import MoleculeContainer, QueryContainer
from ..periodictable import H


def fw_prepare_groups(core: Union[MoleculeContainer, QueryContainer], molecule: MoleculeContainer) -> \
       List[MoleculeContainer]:
    """
    Prepare list of core with connected groups. Hydrogens added to groups for marking connection point.
    Hydrogens have isotope marks equal to mapping of core atoms.
    Groups connected multiple times (rings) - contains multiple hydrogens.

    :param core: core structure for searching
    :param molecule: target structure
    """
    try:
        core_map = next(core.get_mapping(molecule))
    except StopIteration:
        return []

    reverse = {v: k for k, v in core_map.items()}
    cs = set(core_map.values())
    groups = molecule.substructure(molecule._atoms.keys() - cs, recalculate_hydrogens=False)
    gs = set(groups)
    hs = molecule._hydrogens
    hgs = groups._hydrogens
    plane = molecule._plane

    cf = molecule.substructure(cs, recalculate_hydrogens=False)
    chs = cf._hydrogens

    for n, m, b in molecule.bonds():
        if n in cs:
            if m in gs:
                h = H()
                h._Core__isotope = reverse[n]  # mark mapping to isotope
                groups.add_bond(groups.add_atom(h, xy=plane[n]), m, b.copy())
                hgs[m] = hs[m]  # restore H count

                cf.add_bond(cf.add_atom(h.copy(), xy=plane[m]), n, b.copy())
                chs[n] = hs[n]
        elif m in cs and n in gs:
            h = H()
            h._Core__isotope = reverse[m]
            groups.add_bond(groups.add_atom(h, xy=plane[m]), n, b.copy())
            hgs[n] = hs[n]

            cf.add_bond(cf.add_atom(h.copy(), xy=plane[n]), m, b.copy())
            chs[m] = hs[m]
    groups = groups.split()
    groups.insert(0, cf)
    return groups


def fw_decomposition_tree(groups: List[MoleculeContainer]) -> Tree:
    assert len(groups) == len(set(groups))

    pred = {}  # directed graph from substructures to superstructures
    succ = {}
    for m in groups:
        pred[m] = set()
        succ[m] = set()

    for m in groups:
        for n in groups:
            if m < n:
                pred[n].add(m)
                succ[m].add(n)

    # break triangles
    scope = {m for m, ns in succ.items() if len(ns) > 1}
    while scope:
        m = sorted(scope, key=lambda x: len(pred[x]))[0]
        s = succ[m]
        scope.discard(m)
        while True:
            for x, y in product((x for x in s if succ[x]), (x for x in s if len(pred[x]) > 1)):
                if y in succ[x]:
                    s.discard(y)
                    pred[y].discard(m)
                    break
            else:
                break

    def _rec_tree(x):
        return x, [_rec_tree(y) for y in succ[x]]

    m = MoleculeContainer()
    m.add_atom('H')

    return m, [_rec_tree(x) for x, p in pred.items() if not p]


__all__ = ['fw_prepare_groups', 'fw_decomposition_tree']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from chython import smiles, smarts, Transformer
from chython.reactor import deprotection
from itertools import product


def test_deprotection():
    qs = set()
    ts = set()
    for x in dir(deprotection):
        for r in getattr(deprotection, x[9:]):
            if len(r) > 2:  # has test
                q, p, t, a, *bs = r
                t = smiles(t)
                t.canonicalize()
                q = smarts(q)
                qs.add(q)
                ts.add(t)
                a = smiles(a)
                a.canonicalize()
                # test match
                assert q < t, f'{q} !< {t}'
                o = next(Transformer(q, smarts(p))(t))
                assert o == a, f'{o} != {a}'
                for b in bs:
                    b = smiles(b)
                    b.canonicalize()
                    assert not q < b, f'{q} < {b}'

    # test rule-test is unique pair
    assert len(qs) == len(ts)

    m = 0
    for q, t in product(qs, ts):
        m += q < t

    # test selectivity of rules
    assert len(qs) == m



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from ...periodictable import ListElement


def _stripped_rules():
    from ...containers import QueryContainer
    rules = []

    # Ammonia, H-imine,guanidine,amidine. [H][N+]=,-,:
    q = QueryContainer()
    q.add_atom('N', charge=1, hydrogens=(1, 2, 3, 4))
    rules.append(q)
    return rules


def _rules():
    from ...containers import QueryContainer
    rules = _stripped_rules()

    # Phenoles, [H][O,S,Se]-Ar
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), neighbors=1)
    q.add_atom(ListElement(['C', 'N']), hybridization=4)
    q.add_bond(1, 2, 1)
    rules.append(q)

    # Oxo-acids. [H][O,S,Se]-[C,N,P,S,Cl,Se,Br,I]=O
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), neighbors=1)
    q.add_atom(ListElement(['C', 'N', 'P', 'S', 'Cl', 'Se', 'Br', 'I']))
    q.add_atom('O')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # Nitro acid. [H]O-[N+](=O)[O-]
    q = QueryContainer()
    q.add_atom('O', neighbors=1)
    q.add_atom('N', charge=1)
    q.add_atom('O', charge=-1)
    q.add_atom('O')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 2)
    rules.append(q)

    q = QueryContainer()
    q.add_atom(ListElement(['F', 'Cl', 'Br', 'I']), neighbors=0)
    rules.append(q)
    return rules


stripped_rules = Proxy(_stripped_rules)
rules = Proxy(_rules)


__all__ = ['stripped_rules', 'rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Adelia Fatykhova <adelik21979@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import Union
from .base import BaseReactor
from ..containers import QueryContainer, MoleculeContainer


class Transformer(BaseReactor):
    """
    Editor for molecules.
    generates modified molecules from input molecule using template.
    Transformer calling returns generator of all possible replacements.
    """
    def __init__(self, pattern: QueryContainer, replacement: Union[MoleculeContainer, QueryContainer],
                 delete_atoms: bool = True, automorphism_filter: bool = True, fix_aromatic_rings: bool = True,
                 fix_tautomers: bool = True):
        """
        :param pattern: Search pattern.
        :param replacement: Resulted structure.
        :param delete_atoms: If True atoms exists in reactants but not exists in products will be removed.
        :param fix_aromatic_rings: Proceed kekule and thiele on products.
        :param fix_tautomers: See `thiele()` docs.
        :param automorphism_filter: Skip matches to same atoms.
        """
        if not isinstance(pattern, QueryContainer) or not isinstance(replacement, (MoleculeContainer, QueryContainer)):
            raise TypeError('invalid params')

        self.pattern = pattern
        self.replacement = replacement
        self.__automorphism_filter = automorphism_filter
        super().__init__({n for n, h in pattern._masked.items() if not h}, replacement, delete_atoms,
                         fix_aromatic_rings, fix_tautomers)

    def __call__(self, structure: MoleculeContainer):
        if not isinstance(structure, MoleculeContainer):
            raise TypeError('only Molecules possible')

        for mapping in self.pattern.get_mapping(structure, automorphism_filter=self.__automorphism_filter):
            yield from self._patcher(structure, mapping)


__all__ = ['Transformer']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from ...periodictable import ListElement


def _xonyl_groups():
    from ...containers import QueryContainer
    rules = []

    q = QueryContainer()
    q.add_atom('N', neighbors=3, hybridization=2, charge=1)
    q.add_atom('O', neighbors=1, charge=-1)
    q.add_atom('O', neighbors=1)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 2)
    rules.append(q)

    q = QueryContainer()
    q.add_atom(ListElement(['C', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Sb', 'Te', 'I', 'Bi', 'Po', 'At']))
    q.add_atom(ListElement(['S', 'O', 'Se', 'N']), neighbors=1)
    q.add_atom(ListElement(['S', 'O', 'Se', 'N']), neighbors=1)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 2)
    rules.append(q)

    q = QueryContainer()
    q.add_atom(ListElement(['C', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Sb', 'Te', 'I', 'Bi', 'Po', 'At']))
    q.add_atom(ListElement(['S', 'O', 'Se', 'N']), neighbors=1, charge=-1)
    q.add_atom(ListElement(['S', 'O', 'Se', 'N']), neighbors=1)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 2)
    rules.append(q)
    return rules


def _substituents():
    """
    Rules for switchable functional groups remapping
    """
    from ...containers import QueryContainer
    rules = []

    # nitro addition
    #
    #      O [3]     [2] O -- *
    #     //            /:
    # * - N+   >>  * = N+
    #      \            \
    #       O- [2]       O- [3]
    #
    q = QueryContainer()
    q.add_atom('N', neighbors=3, charge=1, hybridization=2)
    q.add_atom('O', neighbors=1, charge=-1)
    q.add_atom('O', neighbors=2)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, (1, 4))
    rules.append((q, ((2, 3), (3, 2))))

    #
    #      NH [3]          NH2 [3]
    #     //               /
    # * - X       >>  * - X
    #      \              :
    #       NH2 [2]       N [2]
    #
    q = QueryContainer()
    q.add_atom('C', neighbors=3)
    q.add_atom('N')
    q.add_atom('N', neighbors=2, hydrogens=0)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, (2, 4))
    rules.append((q, ((2, 3), (3, 2))))

    #
    #      O [3]        [3] O - R
    #     //               /
    # * - X       >>  * - X
    #      \              \\
    #       OH [2]         O [2]
    #
    q = QueryContainer()
    q.add_atom(ListElement(['C', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Sb', 'Te', 'I', 'Bi', 'Po', 'At']))
    q.add_atom(ListElement(['S', 'O', 'Se', 'N']), neighbors=2)
    q.add_atom(ListElement(['S', 'O', 'Se', 'N']), neighbors=1)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 2)
    rules.append((q, ((2, 3), (3, 2))))

    #
    #      O [3]            A
    #     //               /
    # * - X       >>  * - X
    #      \              \\
    #       OH [2]         O [2]
    #
    q = QueryContainer()
    q.add_atom(ListElement(['C', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Sb', 'Te', 'I', 'Bi', 'Po', 'At']))
    q.add_atom('A')
    q.add_atom(ListElement(['S', 'O', 'Se', 'N']), neighbors=1)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 2)
    rules.append((q, ((3, 2),)))  # possible only: (3, 2)
    return rules


xonyl_groups = Proxy(_xonyl_groups)
substituents_groups = Proxy(_substituents)


__all__ = ['xonyl_groups', 'substituents_groups']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from traceback import format_exc
from .emol import EMOLRead
from ...exceptions import EmptyMolecule


class ERXNRead:
    def __init__(self, line, ignore=False, log_buffer=None):
        tmp = line[13:].split()
        self.__reactants_count = int(tmp[0])
        self.__products_count = int(tmp[1])
        self.__reagents_count = int(tmp[2]) if len(tmp) == 3 else 0

        self.__reactants = []
        self.__products = []
        self.__reagents = []
        self.__errors = []
        self.__ignore = ignore
        if log_buffer is None:
            log_buffer = []
        self.__log_buffer = log_buffer

    def __call__(self, line):
        if self.__empty_skip:
            if not line.startswith('M  V30 END CTAB'):
                return
            self.__empty_skip = False
        elif self.__in_mol:
            try:
                x = self.__parser(line)
            except ValueError as e:
                if not self.__ignore:
                    raise
                if not isinstance(e, EmptyMolecule):
                    m = self.__parser._meta  # noqa
                    m['chytorch_molecule_role'] = _roles[self.__parser_group]
                    self.__errors.append(m)
                    self.__log_buffer.append(format_exc())
                    self.__log_buffer.append('bad molecule ignored')
                else:
                    self.__log_buffer.append('empty molecule ignored')
                self.__empty_skip = True
                self.__in_mol -= 1
                if self.__in_mol:
                    self.__parser = EMOLRead(self.__ignore, self.__log_buffer)
            else:
                if x:
                    x = self.__parser.getvalue()
                    self.__in_mol -= 1
                    if self.__in_mol:
                        self.__parser = EMOLRead(self.__ignore, self.__log_buffer)
                    if self.__parser_group == 'REACTANT':
                        self.__reactants.append(x)
                    elif self.__parser_group == 'PRODUCT':
                        self.__products.append(x)
                    elif self.__parser_group == 'AGENT':
                        self.__reagents.append(x)
        elif self.__rend:
            raise ValueError('parser closed')
        elif line.startswith('M  V30 END'):
            if self.__parser_group != line[11:].strip():
                raise ValueError('invalid CTAB')
        elif line.startswith('M  V30 BEGIN'):
            x = line[13:].strip()
            if x == 'REACTANT':
                self.__in_mol = self.__reactants_count
            elif x == 'PRODUCT':
                self.__in_mol = self.__products_count
            elif x == 'AGENT':
                self.__in_mol = self.__reagents_count
            else:
                raise ValueError('invalid RXN CTAB')
            self.__parser_group = x
            if self.__in_mol:
                self.__parser = EMOLRead(self.__ignore, self.__log_buffer)
        elif line.startswith('M  END'):
            self.__rend = True
            return True
        else:
            raise ValueError('invalid CTAB')

    def getvalue(self):
        if self.__rend:
            return {'reactants': self.__reactants, 'products': self.__products, 'reagents': self.__reagents, 'meta': {},
                    'errors': self.__errors}
        raise ValueError('reaction not complete')

    __parser_group = __parser = None
    __rend = __empty_skip = False
    __in_mol = 0


_roles = {'REACTANT': 'reactants', 'PRODUCT': 'products', 'AGENT': 'reagents'}


__all__ = ['ERXNRead']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import deque
from itertools import product
from typing import Iterator, Optional, List
from .. import smarts, ReactionContainer
from .reactor import Reactor, fix_mapping_overlap

"""
Predefined reactors for common reactions.
"""

_amidation = (
    # alerts
    ['[C;z1;x1]-[O;D1]', '[C,N;a]-[O;D1]'],
    # [H,R]COOH + Ar-NH2
    (['[O;M:1]=[C;x2:2][O;D1:3]', '[N;D1:4][C;a;M:5]'], ['[A:2]-[A:4]']),  # 2 reactants, 1 product
    # Alk-NH2
    (['[O;M:1]=[C;x2:2][O;D1:3]', '[N;D1:4][C;z1;x1;M:5]'], ['[A:2]-[A:4]']),
    # Ar-NH-Ar
    (['[O;M:1]=[C;x2:2][O;D1:3]', '[N;D2:4]([C;a;M:5])[C;a;M:6]'], ['[A:2]-[A:4]']),
    # Alk-NH-Ar
    (['[O;M:1]=[C;x2:2][O;D1:3]', '[N;D2:4]([C;a;M:5])[C;z1;x1;M:6]'], ['[A:2]-[A:4]']),
    # Alk2NH
    (['[O;M:1]=[C;x2:2][O;D1:3]', '[N;D2:4]([C;z1;x1;M:5])[C;z1;x1;M:6]'], ['[A:2]-[A:4]']),
    # pyrrole
    (['[O;M:1]=[C;x2:2][O;D1:3]', '[N;a;h1:4]'], ['[A:2]-[A:4]'])
)

_amine_carbonyl_reductive_amination = (
    # alerts
    [],
    # R2C=O + Ar-NH2
    (['[C;x1;z2:1]=[O:2]', '[N;D1:3][C;a;M:4]'], ['[A:1]-[A:3]']),
    # Alk-NH2
    (['[C;x1;z2:1]=[O:2]', '[N;D1:3][C;z1;x1;M:4]'], ['[A:1]-[A:3]']),
    # Ar-NH-Ar
    # (['[C;x1;z2:1]=[O:2]', '[N;D2:3]([C;a;M:4])[C;a;M:5]'], ['[A:1]-[A:3]']),
    # Alk-NH-Ar
    (['[C;x1;z2:1]=[O:2]', '[N;D2:3]([C;a;M:4])[C;z1;x1;M:5]'], ['[A:1]-[A:3]']),
    # Alk2NH
    (['[C;x1;z2:1]=[O:2]', '[N;D2:3]([C;z1;x1;M:4])[C;z1;x1;M:5]'], ['[A:1]-[A:3]'])
)

_suzuki = (
    # alerts
    [],
    # Ar-X + Ar-B
    (['[C;a:1]-[Cl,Br,I;D1:2]', '[C;a:3]-[B;D3;z1:4]'], ['[A:1]-[A:3]']),
    # Ar-X + C=C-B
    (['[C;a:1]-[Cl,Br,I;D1:2]', '[C;x1;z2:3]-[B;D3;z1:4]'], ['[A:1]-[A:3]']),
    # C=C-X + C=C-B
    (['[C;x1;z2:1]-[Cl,Br,I;D1:2]', '[C;x1;z2:3]-[B;D3;z1:4]'], ['[A:1]-[A:3]']),
    # C=C-X + Ar-B
    (['[C;x1;z2:1]-[Cl,Br,I;D1:2]', '[C;a:3]-[B;D3;z1:4]'], ['[A:1]-[A:3]'])
)

_suzuki_amide = (
    # alerts
    [],
    (['[C;a:1]-[B;D3;z1:2]', '[C;a;M:3]-[C:4](=[O;M:5])-N([C;x1;z1])-C(=O)[C;x0;z1]'], ['[A:1]-[A:4]'])
)

_buchwald_hartwig = (
    # alerts
    [],
    # Ar-Hal + Ar-NH2
    (['[C;a:1]-[Cl,Br,I;D1:2]', '[N;D1:3][C;a;M:4]'], ['[A:1]-[A:3]']),
    # Alk-NH2
    (['[C;a:1]-[Cl,Br,I;D1:2]', '[N;D1:3][C;z1;x1;M:4]'], ['[A:1]-[A:3]']),
    # Ar-NH-Ar
    # (['[C;a:1]-[Cl,Br,I;D1:2]', '[N;D2:3]([C;a;M:4])[C;a;M:5]'], ['[A:1]-[A:3]']),
    # Alk-NH-Ar
    (['[C;a:1]-[Cl,Br,I;D1:2]', '[N;D2:3]([C;a;M:4])[C;z1;x1;M:5]'], ['[A:1]-[A:3]']),
    # Alk2NH
    (['[C;a:1]-[Cl,Br,I;D1:2]', '[N;D2:3]([C;z1;x1;M:4])[C;z1;x1;M:5]'], ['[A:1]-[A:3]'])
)

_sulfonamidation = (
    # alerts
    [],
    # RSOOX + Ar-NH2
    (['[O;M:1]=[S:2](=[O;M:3])([C;M:4])-[O,F,Cl,Br,I;D1:5]', '[N;D1:6][C;a;M:7]'], ['[A:2]-[A:6]']),
    # Alk-NH2
    (['[O;M:1]=[S:2](=[O;M:3])([C;M:4])-[O,F,Cl,Br,I;D1:5]', '[N;D1:6][C;z1;x1;M:7]'], ['[A:2]-[A:6]']),
    # Ar-NH-Ar
    (['[O;M:1]=[S:2](=[O;M:3])([C;M:4])-[O,F,Cl,Br,I;D1:5]', '[N;D2:6]([C;a;M:7])[C;a;M:8]'], ['[A:2]-[A:6]']),
    # Alk-NH-Ar
    (['[O;M:1]=[S:2](=[O;M:3])([C;M:4])-[O,F,Cl,Br,I;D1:5]', '[N;D2:6]([C;a;M:7])[C;z1;x1;M:8]'], ['[A:2]-[A:6]']),
    # Alk2NH
    (['[O;M:1]=[S:2](=[O;M:3])([C;M:4])-[O,F,Cl,Br,I;D1:5]', '[N;D2:6]([C;z1;x1;M:7])[C;z1;x1;M:8]'], ['[A:2]-[A:6]']),
    (['[O;M:1]=[S:2](=[O;M:3])([C;M:4])-[O,F,Cl,Br,I;D1:5]', '[N;a;h1:6]'], ['[A:2]-[A:6]'])
)

_amine_isocyanate = (
    # alerts
    [],
    # RN=C=O + Ar-NH2
    (['[C;M:1][N:2]=[C:3]=[O;M:4]', '[N;D1:5][C;a;M:6]'], ['[A:2][A:3]-[A:5]']),
    # Alk-NH2
    (['[C;M:1][N:2]=[C:3]=[O;M:4]', '[N;D1:5][C;z1;x1;M:6]'], ['[A:2][A:3]-[A:5]']),
    # Ar-NH-Ar
    (['[C;M:1][N:2]=[C:3]=[O;M:4]', '[N;D2:5]([C;a;M:6])[C;a;M:7]'], ['[A:2][A:3]-[A:5]']),
    # Alk-NH-Ar
    (['[C;M:1][N:2]=[C:3]=[O;M:4]', '[N;D2:5]([C;a;M:6])[C;z1;x1;M:7]'], ['[A:2][A:3]-[A:5]']),
    # Alk2NH
    (['[C;M:1][N:2]=[C:3]=[O;M:4]', '[N;D2:5]([C;z1;x1;M:6])[C;z1;x1;M:7]'], ['[A:2][A:3]-[A:5]'])
)

_sonogashira = (
    # alerts
    [],
    # Ar-Hal + HC#C-R
    (['[C;a:1]-[Cl,Br,I;D1:2]', '[C;D1:3]#[C;D2;M:4]'], ['[A:1]-[A:3]']),
    # C=C-Hal
    (['[C;x1;z2:1]-[Cl,Br,I;D1:2]', '[C;D1:3]#[C;D2;M:4]'], ['[A:1]-[A:3]']),
    # R-C(=O)-Hal
    (['[C;x2:1](=[O;M:2])-[Cl,Br,I;D1:3]', '[C;D1:4]#[C;D2;M:5]'], ['[A:1]-[A:4]'])
)


#################
# Magic Factory #
#################

__all__ = [k[1:] for k, v in globals().items() if k.startswith('_') and isinstance(v, tuple) and v]

_cache = {}


def _prepare_reactor(rules, name):
    rxn = [Reactor([smarts(r) for r in rs], [smarts(p) for p in ps], one_shot=False, automorphism_filter=False)  # noqa
           for rs, ps in rules[1:]]
    rxn_os = [Reactor([smarts(r) for r in rs], [smarts(p) for p in ps], automorphism_filter=False)  # noqa
              for rs, ps in rules[1:]]
    alerts = [smarts(x) for x in rules[0]]

    def w(*molecules, one_shot=False, rules_set: Optional[List[int]] = None,
          alerts_set: Optional[List[int]] = None, excess: Optional[List[int]] = None) -> Iterator[ReactionContainer]:
        """
        Generate reactions for given reactants.

        Rules:

        %s

        Alerts:

        %s

        :param molecules: Reactants molecules.
        :param one_shot: Generate only single stage products. Otherwise, all possible combinations, including products.
        :param rules_set: Select only specific rules.
        :param alerts_set: Check only specific structural alerts of reactants. Pass empty list to disable any checks.
        :param excess: Molecules indices which can be involved in multistep synthesis. All by default.
        """
        if not molecules:
            raise ValueError('empty molecule list')
        alerts_set = alerts if alerts_set is None else [alerts[x] for x in alerts_set]
        if any(a < m for a, m in product(alerts_set, molecules)):
            return

        molecules = fix_mapping_overlap(molecules)
        seen = set()
        if one_shot:
            picked_rxn = enumerate(rxn_os) if rules_set is None else [(x, rxn_os[x]) for x in rules_set]
            for i, rx in picked_rxn:
                for r in rx(*molecules):
                    if str(r) in seen:
                        continue
                    seen.add(str(r))
                    r.meta['MATCHED_RULES'] = [i]
                    yield r
            return

        excess = molecules if excess is None else [molecules[x] for x in excess]
        picked_rxn = list(enumerate(rxn)) if rules_set is None else [(x, rxn[x]) for x in rules_set]

        stack = deque([])
        for i, r in enumerate(picked_rxn):
            x = picked_rxn.copy()
            del x[i]
            stack.appendleft((r, [], molecules, x))

        while stack:
            (i, rx), chain, rct, nxt_rxn = stack.pop()
            chain = chain + [i]
            for r in rx(*rct):
                if str(r) in seen:
                    continue
                seen.add(str(r))

                r = ReactionContainer([x.copy() for x in molecules], r.products)
                r.meta['MATCHED_RULES'] = chain
                yield r

                x = excess.copy()
                for p in reversed(r.products):
                    x.insert(0, p.copy())
                x = fix_mapping_overlap(x)
                if excess is not molecules:
                    # expected that product can react with all excess molecules simultaneously.
                    # e.g. multicomponent reaction (Ugi)
                    for m, nrx in enumerate(nxt_rxn):
                        z = nxt_rxn.copy()
                        del z[m]
                        stack.append((nrx, chain, x.copy(), z))
                else:  # drop one of the reactants
                    for n in range(len(r.products), len(x)):
                        y = x.copy()
                        del y[n]
                        for m, nrx in enumerate(nxt_rxn):
                            z = nxt_rxn.copy()
                            del z[m]
                            stack.append((nrx, chain, y, z))

    w.__module__ = __name__
    w.__qualname__ = w.__name__ = name
    doc = w.__doc__ % ('\n'.join(f'{n}: {".".join(r)}>>{".".join(p)}' for n, (r, p) in enumerate(rules[1:])),
                       '\n'.join(f'{n}: {r}' for n, r in enumerate(rules[0])) or '--')
    w.__doc__ = '\n'.join(x.lstrip() for x in doc.splitlines())
    return w


def __getattr__(name):
    try:
        return _cache[name]
    except KeyError:
        if name in __all__:
            _cache[name] = t = _prepare_reactor(globals()[f'_{name}'], name)
            return t
        raise AttributeError


def __dir__():
    return __all__



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .element import Element
from .groups import GroupI
from .periods import *


class H(Element, PeriodI, GroupI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 1

    @property
    def isotopes_distribution(self):
        return {1: 0.999885, 2: 0.000115, 3: 0.}

    @property
    def isotopes_masses(self):
        return {1: 1.007825, 2: 2.014102, 3: 3.016049}

    @property
    def _common_valences(self):
        return 1,

    @property
    def _valences_exceptions(self):
        return (1, False, 0, ()), (0, True, 0, ()), (-1, False, 0, ())

    @property
    def atomic_radius(self):
        return 0.53


class Li(Element, PeriodII, GroupI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 3

    @property
    def isotopes_distribution(self):
        return {6: 0.0759, 7: 0.9241}

    @property
    def isotopes_masses(self):
        return {6: 6.015122, 7: 7.016004}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return (1, False, 0, ()),

    @property
    def atomic_radius(self):
        return 167


class Na(Element, PeriodIII, GroupI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 11

    @property
    def isotopes_distribution(self):
        return {22: 0., 23: 1.0}

    @property
    def isotopes_masses(self):
        return {22: 21.994437, 23: 22.98977}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return (1, False, 0, ()),

    @property
    def atomic_radius(self):
        return 1.9


class K(Element, PeriodIV, GroupI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 19

    @property
    def isotopes_distribution(self):
        return {39: 0.932581, 40: 0.000117, 41: 0.067302, 42: 0.}

    @property
    def isotopes_masses(self):
        return {39: 38.963707, 40: 39.963999, 41: 40.961826, 42: 41.962402}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return (1, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.43


class Rb(Element, PeriodV, GroupI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 37

    @property
    def isotopes_distribution(self):
        return {82: 0., 85: 0.7217, 87: 0.2783}

    @property
    def isotopes_masses(self):
        return {82: 81.918209, 85: 84.911789, 87: 86.909183}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return (1, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.65


class Cs(Element, PeriodVI, GroupI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 55

    @property
    def isotopes_distribution(self):
        return {131: 0., 133: 1.0}

    @property
    def isotopes_masses(self):
        return {131: 130.905464, 133: 132.905447}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return (1, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.98


class Fr(Element, PeriodVII, GroupI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 87

    @property
    def isotopes_distribution(self):
        return {223: 1.0}

    @property
    def isotopes_masses(self):
        return {223: 223.019736}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return (1, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.98  # unknown, taken radius of previous element in group


__all__ = ['H', 'Li', 'Na', 'K', 'Rb', 'Cs', 'Fr']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Aleksandr Sizov <murkyrussian@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict, deque
from math import log2
from numpy import uint8, zeros
from typing import Deque, Dict, List, Set, Tuple, TYPE_CHECKING


if TYPE_CHECKING:
    from chython import MoleculeContainer


class LinearFingerprint:
    __slots__ = ()
    """
    Linear fragments fingerprints.
    Transform structures into fingerprints based on linear fragments descriptors.
    Also count of fragments takes into account by activating multiple bits, but less or equal to `number_bit_pairs`.

    For example `CC` fragment found 4 times and `number_bit_pairs` is set to 3.
    In this case will be activated 3 bits: for count 1, for count 2 and for count 3.
    This gives intersection in bits with another structure with only 2 `CC` fragments.
    """
    def linear_fingerprint(self, min_radius: int = 1, max_radius: int = 4,
                           length: int = 1024, number_active_bits: int = 2, number_bit_pairs: int = 4):
        """
        Transform structures into array of binary features.

        :param min_radius: minimal length of fragments
        :param max_radius: maximum length of fragments
        :param length: bit string's length. Should be power of 2
        :param number_active_bits: number of active bits for each hashed tuple
        :param number_bit_pairs: describe how much repeating fragments we can count in hashable
               fingerprint (if number of fragment in molecule greater or equal this number, we will be
               activate only this number of fragments

        :return: array(n_features)
        """
        bits = self.linear_bit_set(min_radius, max_radius, length, number_active_bits, number_bit_pairs)
        fingerprints = zeros(length, dtype=uint8)
        fingerprints[list(bits)] = 1
        return fingerprints

    def linear_bit_set(self, min_radius: int = 1, max_radius: int = 4,
                       length: int = 1024, number_active_bits: int = 2, number_bit_pairs: int = 4) -> Set[int]:
        """
        Transform structure into set of indexes of True-valued features.

        :param min_radius: minimal length of fragments
        :param max_radius: maximum length of fragments
        :param length: bit string's length. Should be power of 2
        :param number_active_bits: number of active bits for each hashed tuple
        :param number_bit_pairs: describe how much repeating fragments we can count in hashable
               fingerprint (if number of fragment in molecule greater or equal this number, we will be
               activate only this number of fragments
        """
        mask = length - 1
        log = int(log2(length))

        hashes = self.linear_hash_set(min_radius, max_radius, number_bit_pairs)
        active_bits = set()
        for tpl in hashes:
            active_bits.add(tpl & mask)
            if number_active_bits == 2:
                active_bits.add((tpl >> log) & mask)
            elif number_active_bits > 2:
                for _ in range(1, number_active_bits):
                    tpl >>= log  # shift
                    active_bits.add(tpl & mask)
        return active_bits

    def linear_hash_set(self, min_radius: int = 1, max_radius: int = 4,
                        number_bit_pairs: int = 4) -> Set[int]:
        """
        Transform structure into set of integer hashes of fragments with count information.

        :param min_radius: minimal length of fragments
        :param max_radius: maximum length of fragments
        :param number_bit_pairs: describe how much repeating fragments we can count in hashable
               fingerprint (if number of fragment in molecule greater or equal this number, we will be
               activate only this number of fragments
        """
        return {hash((*tpl, cnt)) for tpl, count in
                self._fragments(min_radius, max_radius).items()
                for cnt in range(min(len(count), number_bit_pairs))}

    def _chains(self: 'MoleculeContainer', min_radius: int = 1, max_radius: int = 4) -> Set[Tuple[int, ...]]:
        queue: Deque[Tuple[int, ...]]  # typing
        atoms = self._atoms
        bonds = self._bonds

        if min_radius == 1:
            arr = {(x,) for x in atoms}
            if max_radius == 1:  # special case
                return arr
            else:
                queue = deque(arr)
        else:
            arr = set()
            queue = deque((x,) for x in atoms)

        while queue:
            now = queue.popleft()
            var = [now + (x,) for x in bonds[now[-1]] if x not in now]
            if var:
                if len(var[0]) < max_radius:
                    queue.extend(var)
                if len(var[0]) >= min_radius:
                    for frag in var:
                        rev = frag[::-1]
                        arr.add(frag if frag > rev else rev)
        return arr

    def _fragments(self: 'MoleculeContainer', min_radius: int = 1, max_radius: int = 4) -> Dict[Tuple[int, ...],
                                                                                                List[Tuple[int, ...]]]:
        atoms = {idx: hash((atom.isotope or 0, atom.atomic_number, atom.charge, atom.is_radical))
                 for idx, atom in self.atoms()}

        bonds = self._bonds
        out = defaultdict(list)

        for frag in self._chains(min_radius, max_radius):
            var = [atoms[frag[0]]]
            for x, y in zip(frag, frag[1:]):
                var.append(int(bonds[x][y]))
                var.append(atoms[y])
            var = tuple(var)
            rev_var = var[::-1]
            if var > rev_var:
                out[var].append(frag)
            else:
                out[rev_var].append(frag)
        return dict(out)


__all__ = ['LinearFingerprint']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_method
from functools import reduce
from hashlib import sha512
from itertools import chain
from math import ceil
from operator import itemgetter, or_
from typing import Dict, Iterable, Iterator, Optional, Tuple, List
from zlib import compress, decompress
from .cgr import CGRContainer
from .molecule import MoleculeContainer
from ..algorithms.calculate2d import Calculate2DReaction
from ..algorithms.depict import DepictReaction
from ..algorithms.mapping import Mapping
from ..algorithms.standardize import StandardizeReaction


class ReactionContainer(StandardizeReaction, Mapping, Calculate2DReaction, DepictReaction):
    """
    Reaction storage. Contains reactants, products and reagents lists.

    Reaction storage hashable and comparable. based on reaction unique signature (SMILES).
    """
    __slots__ = ('__reactants', '__products', '__reagents', '__meta', '__name', '_arrow', '_signs', '__dict__')
    __class_cache__ = {}

    def __init__(self, reactants: Iterable[MoleculeContainer] = (), products: Iterable[MoleculeContainer] = (),
                 reagents: Iterable[MoleculeContainer] = (), meta: Optional[Dict] = None, name: Optional[str] = None):
        """
        New reaction object creation

        :param reactants: list of MoleculeContainers in left side of reaction
        :param products: right side of reaction. see reactants
        :param reagents: middle side of reaction: solvents, catalysts, etc. see reactants
        :param meta: dictionary of metadata. like DTYPE-DATUM in RDF

        """
        reactants = tuple(reactants)
        products = tuple(products)
        reagents = tuple(reagents)
        if not reactants and not products and not reagents:
            raise ValueError('At least one graph object required')
        elif not all(isinstance(x, MoleculeContainer) for x in chain(reactants, products, reagents)):
            raise TypeError(f'MoleculeContainers expected')

        self.__reactants = reactants
        self.__products = products
        self.__reagents = reagents
        if meta is None:
            self.__meta = None
        else:
            self.__meta = dict(meta)
        if name is None:
            self.__name = None
        else:
            self.name = name
        self._arrow = None
        self._signs = None

    @property
    def reactants(self) -> Tuple[MoleculeContainer, ...]:
        return self.__reactants

    @property
    def reagents(self) -> Tuple[MoleculeContainer, ...]:
        return self.__reagents

    @property
    def products(self) -> Tuple[MoleculeContainer, ...]:
        return self.__products

    def molecules(self) -> Iterator[MoleculeContainer]:
        """
        Iterator of all reaction molecules
        """
        return chain(self.__reactants, self.__reagents, self.__products)

    @property
    def meta(self) -> Dict:
        """
        Dictionary of metadata.
        Like DTYPE-DATUM in RDF
        """
        if self.__meta is None:
            self.__meta = {}  # lazy
        return self.__meta

    @property
    def name(self) -> str:
        return self.__name or ''

    @name.setter
    def name(self, name: str):
        if not isinstance(name, str):
            raise TypeError('name should be string up to 80 symbols')
        self.__name = name

    def copy(self) -> 'ReactionContainer':
        """
        Get copy of object
        """
        copy = object.__new__(self.__class__)
        copy._ReactionContainer__reactants = tuple(x.copy() for x in self.__reactants)
        copy._ReactionContainer__products = tuple(x.copy() for x in self.__products)
        copy._ReactionContainer__reagents = tuple(x.copy() for x in self.__reagents)
        copy._ReactionContainer__name = self.__name
        if self.__meta is None:
            copy._ReactionContainer__meta = None
        else:
            copy._ReactionContainer__meta = self.__meta.copy()
        copy._arrow = self._arrow
        copy._signs = self._signs
        return copy

    @cached_method
    def compose(self) -> CGRContainer:
        """
        Get CGR of reaction

        Reagents will be presented as unchanged molecules
        :return: CGRContainer
        """
        rr = self.__reagents + self.__reactants
        if rr:
            r = reduce(or_, rr)
        else:
            r = MoleculeContainer()
        if self.__products:
            p = reduce(or_, self.__products)
        else:
            p = MoleculeContainer()
        return r ^ p

    def flush_cache(self):
        self.__dict__.clear()
        for m in self.molecules():
            m.flush_cache()

    def pack(self, *, compressed=True, check=True):
        """
        Pack into compressed bytes.

        Note:
            * Same restrictions as in molecules pack.
            * reactants, reagents nad products should contain less than 256 molecules.

        Format specification:
        Big endian bytes order
        8 bit - header byte = 0x01 (current format specification)
        8 bit - reactants count
        8 bit - reagents count
        8 bit - products count
        x bit - concatenated molecules packs

        :param compressed: return zlib-compressed pack.
        :param check: check molecules for format restrictions.
        """
        data = b''.join((bytearray((1, len(self.__reactants), len(self.__reagents), len(self.__products))),
                         *(m.pack(compressed=False, check=check) for m in self.molecules())))
        if compressed:
            return compress(data, 9)
        return data

    @classmethod
    def pack_len(cls, data: bytes, /, *, compressed=True) -> Tuple[List[int], List[int], List[int]]:
        """
        Returns reactants, reagents, products molecules atoms count in reaction pack.
        """
        if compressed:
            data = decompress(data)
        data = memoryview(data)
        if data[0] != 1:
            raise ValueError('invalid pack header')
        reactants, reagents, products = data[1], data[2], data[3]

        v = data[4]  # mol pack version
        shift = 5  # RH+RC+RC+PC+MH
        molecules = []
        for _ in range(reactants + reagents + products - 1):
            acs = int.from_bytes(data[shift: shift + 3], 'big')
            neighbors = 0
            ac = acs >> 12
            shift += 4  # AC+CC+AN
            for _ in range(ac):
                neighbors += data[shift] & 0x0f
                shift += 9
            neighbors //= 2
            if v == 2:
                shift += 3 * neighbors + ceil(neighbors * 3 / 8) + (acs & 0x0fff) * 4
            elif v == 0:
                shift += 3 * neighbors + ceil(neighbors / 5) * 2 + (acs & 0x0fff) * 4
            molecules.append(ac)
        if reactants or reagents or products:
            molecules.append(int.from_bytes(data[shift: shift + 3], 'big') >> 12)
        return molecules[:reactants], molecules[reactants: -products], molecules[-products:]

    @classmethod
    def unpack(cls, data: bytes, /, *, compressed=True) -> 'ReactionContainer':
        """
        Unpack from compressed bytes.

        :param compressed: decompress data before processing.
        """
        if compressed:
            data = decompress(data)
        data = memoryview(data)
        if data[0] != 1:
            raise ValueError('invalid pack header')

        reactants, reagents, products = data[1], data[2], data[3]
        molecules = []
        shift = 4
        for _ in range(reactants + reagents + products):
            m, pl = MoleculeContainer.unpack(data[shift:], compressed=False, _return_pack_length=True)
            molecules.append(m)
            shift += pl
        return cls(molecules[:reactants], molecules[-products:], molecules[reactants: -products])

    def __invert__(self) -> CGRContainer:
        """
        Get CGR of reaction
        """
        return self.compose()

    def __eq__(self, other):
        return isinstance(other, ReactionContainer) and str(self) == str(other)

    @cached_method
    def __hash__(self):
        return hash(str(self))

    @cached_method
    def __bytes__(self):
        return sha512(str(self).encode()).digest()

    def __bool__(self):
        """
        Exists both reactants and products
        """
        return bool(self.__reactants and self.__products)

    @cached_method
    def __str__(self):
        return format(self)

    def __format__(self, format_spec):
        """
        :param format_spec:
            !c - Keep nested containers order.
            a - Generate asymmetric closures.
            !s - Disable stereo marks.
            A - Use aromatic bonds instead aromatic atoms.
            m - Set atom mapping.
            r - Generate random-ordered smiles.
            h - Show implicit hydrogens.
            !b - Disable bonds tokens.
            !x - Disable CXSMILES extension.
            !z - Disable charge representation.
        """
        sig = []
        count = 0
        contract = []
        orders = []

        for ml in (self.__reactants, self.__reagents, self.__products):
            mso = [(m, *m.__format__(format_spec, _return_order=True)) for m in ml]
            if not format_spec or '!c' not in format_spec:
                mso.sort(key=itemgetter(1))

            ss = []
            for m, s, o in mso:
                if m.connected_components_count > 1:
                    contract.append([str(x + count) for x in range(m.connected_components_count)])
                    count += m.connected_components_count
                else:
                    count += 1

                orders.append((m, o))
                ss.append(s)
            sig.append('.'.join(ss))

        if not format_spec or '!x' not in format_spec:
            cx = []
            if r := ','.join(str(n) for n, (m, a) in enumerate((m, a) for m, o in orders for a in o) if m._radicals[a]):
                cx.append(f'^1:{r}')
            if contract:
                cx.append(f"f:{','.join('.'.join(x) for x in contract)}")
            if cx:
                return f"{'>'.join(sig)} |{','.join(cx)}|"
        return '>'.join(sig)

    @cached_method
    def __len__(self):
        return len(self.__reactants) + len(self.__products) + len(self.__reagents)

    def __getstate__(self):
        state = {'reactants': self.__reactants, 'products': self.__products, 'reagents': self.__reagents,
                 'meta': self.__meta, 'name': self.__name, 'arrow': self._arrow, 'signs': self._signs}
        from chython import pickle_cache

        if pickle_cache:
            state['cache'] = self.__dict__
        return state

    def __setstate__(self, state):
        self.__reactants = state['reactants']
        self.__products = state['products']
        self.__reagents = state['reagents']
        self.__meta = state['meta']
        self.__name = state['name']
        self._arrow = state['arrow']
        self._signs = state['signs']
        if 'cache' in state:
            self.__dict__.update(state['cache'])


__all__ = ['ReactionContainer']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  Copyright 2014-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from distutils.command.sdist import sdist
from distutils.command.build import build
from distutils.util import get_platform
from importlib.util import find_spec
from pathlib import Path
from setuptools import setup, Extension, find_packages


class _sdist(sdist):
    def finalize_options(self):
        super().finalize_options()
        self.distribution.data_files.append(('lib', ['INCHI/libinchi.so',
                                                     'INCHI/libinchi.dll', 'INCHI/libinchi.dynlib']))


cmd_class = {'sdist': _sdist}


if find_spec('wheel'):
    from wheel.bdist_wheel import bdist_wheel

    class _bdist_wheel(bdist_wheel):
        def finalize_options(self):
            super().finalize_options()
            self.root_is_pure = False
            platform = get_platform()
            if platform == 'win-amd64':
                self.distribution.data_files.append(('lib', ['INCHI/libinchi.dll']))
            elif platform == 'linux-x86_64':
                self.distribution.data_files.append(('lib', ['INCHI/libinchi.so']))
            elif platform == 'macosx-11-x86_64':
                self.distribution.data_files.append(('lib', ['INCHI/libinchi.dynlib']))

    cmd_class['bdist_wheel'] = _bdist_wheel


if find_spec('cython'):
    class _build(build):
        def finalize_options(self):
            super().finalize_options()
            from Cython.Build import cythonize
            self.distribution.ext_modules = cythonize(self.distribution.ext_modules, language_level=3)

    cmd_class['build'] = _build


setup(
    name='chython',
    version='1.48.1',
    packages=find_packages(),
    url='https://github.com/chython/chython',
    license='LGPLv3',
    author='Dr. Ramil Nugmanov',
    author_email='nougmanoff@protonmail.com',
    python_requires='>=3.8',
    cmdclass=cmd_class,
    ext_modules=[Extension('chython.algorithms._isomorphism', ['chython/algorithms/_isomorphism.pyx'],
                           extra_compile_args=['-O3']),
                 Extension('chython.containers._unpack', ['chython/containers/_unpack.pyx'],
                           extra_compile_args=['-O3']),
                 Extension('chython.containers._pack', ['chython/containers/_pack.pyx'], extra_compile_args=['-O3']),
                 Extension('chython.files._xyz', ['chython/files/_xyz.pyx'], extra_compile_args=['-O3'])],
    setup_requires=['wheel', 'cython'],
    install_requires=['CachedMethods>=0.1.4,<0.2', 'lazy-object-proxy>=1.6', 'lxml>=4.1', 'py-mini-racer>=0.4.0',
                      'numpy>=1.18'],
    extras_require={'pytest': ['pytest'], 'mapping': ['chytorch-rxnmap>=1.0']},
    package_data={'chython.algorithms': ['_isomorphism.pyx'], 'chython.algorithms.calculate2d': ['clean2d.js'],
                  'chython.containers': ['_pack.pyx', '_unpack.pyx'], 'chython.files': ['_xyz.pyx']},
    data_files=[],
    zip_safe=False,
    long_description=(Path(__file__).parent / 'README.rst').read_text('utf8'),
    classifiers=['Environment :: Plugins',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 3 :: Only',
                 'Programming Language :: Python :: 3.8',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Chemistry',
                 'Topic :: Scientific/Engineering :: Information Analysis',
                 'Topic :: Software Development',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: Software Development :: Libraries :: Python Modules']
)



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import class_cached_property
from itertools import chain, count, repeat
from logging import getLogger, INFO
from numpy import ix_, unravel_index, argmax, zeros, array, isclose, nonzero, ones, mean
from typing import TYPE_CHECKING, Union


if TYPE_CHECKING:
    from chython import ReactionContainer

logger = getLogger('chython.attention')
logger.setLevel(INFO)


class Attention:
    __slots__ = ()

    def reset_mapping(self: Union['ReactionContainer', 'Attention'], *, return_score: bool = False, multiplier=1.75,
                      keep_reactants_numbering=False) -> Union[bool, float]:
        """
        Do atom-to-atom mapping. Return True if mapping changed.
        """
        if any(len(bs) > 14 for m in self.molecules() for bs in m._bonds.values()):
            logger.info('atom-to-atom mapping not supported for hypervalent compounds')
            return False
        fixed = self.__fix_collisions()
        equal_atoms, p2r, r2p, r_adj, p_adj, r_map, p_map, pa = self.__prepare_remapping()

        # rxnmapper-inspired algorithm
        am = self.__get_attention()
        # sum of reactants to products attention and vice-versa for equal atom types only
        am = (am[p2r] + am[r2p].T) * equal_atoms
        amc = am.copy()

        mapping = {}
        scope = zeros(pa, dtype=bool)
        seen = ones(pa, dtype=bool)
        score = []
        for x in range(pa):  # iteratively map each product atom to reactant
            # select highest attention
            # todo: optimize
            if not x:
                i, j = unravel_index(argmax(am), am.shape)
            else:
                ams = am[scope]
                if ams.size:
                    i, j = unravel_index(argmax(ams), ams.shape)
                    i = nonzero(scope)[0][i]
                else:
                    i, j = unravel_index(argmax(am), am.shape)
            if isclose(am[i, j], 0.):  # no more products atoms in reactants
                # mark as unmapped
                for n in set(p_map).difference(mapping):
                    mapping[n] = 0
                break
            else:
                score.append(amc[i, j])
            mapping[p_map[i]] = r_map[j]
            am[ix_(p_adj[i], r_adj[j])] *= multiplier  # highlight neighbors
            am[i] = am[:, j] = 0  # mask mapped product and reactant atoms
            seen[i] = False
            scope[i] = False
            scope[p_adj[i] & seen] = True

        score = float(mean(score)) if score else 0.
        # mapping done.
        if any(n != m for n, m in mapping.items()):  # old mapping changed
            if keep_reactants_numbering or fixed:
                r_mapping = {n: n for n in r_map}
            else:
                r_mapping = {m: n for n, m in enumerate(r_map, 1)}  # remap reactants to contiguous range
                for m in self.reactants:
                    m.remap(r_mapping)

            p_mapping = {}
            nm = max(r_mapping.values())
            for n, m in mapping.items():
                if m := r_mapping.get(m):
                    p_mapping[n] = m
                else:  # not found in reactants atoms. set unique numbers.
                    nm += 1
                    p_mapping[n] = nm

            for m in self.products:
                m.remap(p_mapping)
            self.flush_cache()
            fixed = True

        if self.fix_groups_mapping():  # fix carboxy etc
            fixed = True
        if self.fix_mapping():  # fix common mistakes in mechanisms
            fixed = True
        if return_score:
            return score
        return fixed

    def __fix_collisions(self: 'ReactionContainer'):
        r = [n for m in chain(self.reactants, self.reagents) for n in m._atoms]
        p = [n for m in self.products for n in m._atoms]
        c = count(1)
        if len(r) != len(set(r)):
            for m in chain(self.reactants, self.reagents):
                m.remap({n: next(c) for n in m._atoms})
        if len(p) != len(set(p)):
            for m in self.products:
                m.remap({n: next(c) for n in m._atoms})
        if next(c) != 1:
            self.flush_cache()
            return True
        return False

    def __prepare_remapping(self: 'ReactionContainer'):
        r_map = [n for m in self.reactants for n in m]
        p_map = [n for m in self.products for n in m]
        ra = len(r_map)  # number of reactants atoms
        pa = len(p_map)  # number of products atoms

        ram = [False]  # reactants atoms mask
        r_atoms = []
        r_adj = zeros((ra, ra), dtype=bool)
        i = 0
        for m in self.reactants:
            ram.append(False)
            ram.extend(repeat(True, len(m)))
            a = m.adjacency_matrix()
            j = i + len(m)
            r_adj[i:j, i:j] = a
            i = j
            r_atoms.extend(a.atomic_number for _, a in m.atoms())
        r_atoms = array(r_atoms, dtype=int)

        pam = [False] * len(ram)  # products atoms mask
        p_atoms = []
        p_adj = zeros((pa, pa), dtype=bool)
        i = 0
        for m in self.products:
            pam.append(False)
            pam.extend(repeat(True, len(m)))
            a = m.adjacency_matrix()
            j = i + len(m)
            p_adj[i:j, i:j] = a
            i = j
            p_atoms.extend(a.atomic_number for _, a in m.atoms())
        p_atoms = array(p_atoms, dtype=int)

        ram.extend(repeat(False, len(pam) - len(ram)))
        ram = array(ram, dtype=bool)
        pam = array(pam, dtype=bool)
        return p_atoms[:, None] == r_atoms, ix_(pam, ram), ix_(ram, pam), r_adj, p_adj, r_map, p_map, pa

    @class_cached_property
    def __attention_model(self):
        from chython import torch_device
        from chytorch.zoo.rxnmap import Model

        model = Model.pretrained().to(torch_device)
        for p in model.parameters():
            p.requires_grad = False
        return model

    @class_cached_property
    def __cutoff(self):
        return self.__attention_model.encoder.molecule_encoder.spatial_encoder.num_embeddings - 3

    @class_cached_property
    def __autocast(self):
        from chython import torch_device

        if torch_device.startswith('cuda'):
            try:
                from torch import autocast
            except ImportError:  # torch 1.8 ad-hoc
                from torch.cuda.amp import autocast

                return autocast()
            else:
                return autocast('cuda')
        return autocast_filler()

    def __get_attention(self):
        from chython import torch_device
        from torch import no_grad
        from chytorch.utils.data import ReactionDataset

        b = [x.unsqueeze(0).to(torch_device) for x in ReactionDataset([self], distance_cutoff=self.__cutoff)[0]]
        with no_grad(), self.__autocast:
            am = self.__attention_model(b, mapping_task=True).squeeze(0).float().cpu().numpy()
        return am


class autocast_filler:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        ...


__all__ = ['Attention']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from ctypes import c_char, c_double, c_short, c_long, c_char_p, c_byte, POINTER, Structure, cdll, byref
from distutils.util import get_platform
from fileinput import FileInput
from io import StringIO, TextIOWrapper
from itertools import count
from os import name
from pathlib import Path
from re import split
from sys import prefix, exec_prefix
from typing import List, Iterator
from warnings import warn
from ._mdl import Parser, common_isotopes
from ..containers import MoleculeContainer
from ..containers.bonds import Bond
from ..exceptions import ValenceError, IsChiral, NotChiral, ParseError
from ..periodictable import H


class INCHIRead(Parser):
    """
    INCHI separated per lines files reader. works similar to opened file object. support `with` context manager.
    on initialization accept opened in text mode file, string path to file,
    pathlib.Path object or another buffered reader object.
    line should be start with INCHI string and
    optionally continues with space/tab separated list of key:value [or key=value] data if header=None.

        example:
            InChI=1S/C2H5/c1-2/h1H2,2H3/q+1 id:123 key=value

    if header=True then first line of file should be space/tab separated list of keys including INCHI column key.

        example:
            ignored_inchi_key key1 key2
            InChI=1S/C2H5/c1-2/h1H2,2H3/q+1 1 2

    also possible to pass list of keys (without inchi_pseudo_key) for mapping space/tab separated list
    of INCHI and values: header=['key1', 'key2'] # order depended
    """
    def __init__(self, file, header=None, ignore_stereo=False, **kwargs):
        """
        :param ignore: Skip some checks of data or try to fix some errors.
        :param remap: Remap atom numbers started from one.
        :param store_log: Store parser log if exists messages to `.meta` by key `ParserLog`.
        :param calc_cis_trans: Calculate cis/trans marks from 2d coordinates.
        :param ignore_stereo: Ignore stereo data.
        :param ignore_bad_isotopes: reset invalid isotope mark to non-isotopic.
        """
        if isinstance(file, str):
            self._file = open(file)
            self.__is_buffer = False
        elif isinstance(file, Path):
            self._file = file.open()
            self.__is_buffer = False
        elif isinstance(file, (TextIOWrapper, StringIO, FileInput)):
            self._file = file
            self.__is_buffer = True
        else:
            raise TypeError('invalid file. TextIOWrapper, StringIO subclasses possible')
        super().__init__(**kwargs)
        self.__file = iter(self._file.readline, '')

        if header is True:
            self.__header = next(self.__file).split()[1:]
        elif header:
            if not isinstance(header, (list, tuple)) or not all(isinstance(x, str) for x in header):
                raise TypeError('expected list (tuple) of strings')
            self.__header = header
        else:
            self.__header = None

        self.__ignore_stereo = ignore_stereo
        self._data = self.__data()

    def __data(self):
        file = self._file
        parse = self.parse
        try:
            seekable = file.seekable()
        except AttributeError:
            seekable = False
        pos = file.tell() if seekable else None
        for n, line in enumerate(self.__file):
            try:
                x = parse(line)
            except ValueError:
                yield ParseError(n, pos, self._format_log(), line)
            else:
                yield x
            if seekable:
                pos = file.tell()

    @classmethod
    def create_parser(cls, header=None, ignore_stereo=False, *args, **kwargs):
        """
        Create INCHI parser function configured same as INCHIRead object
        """
        obj = object.__new__(cls)
        obj._INCHIRead__header = header
        obj._INCHIRead__ignore_stereo = ignore_stereo
        super(INCHIRead, obj).__init__(*args, **kwargs)
        return obj.parse

    def close(self, force=False):
        """
        close opened file

        :param force: force closing of externally opened file or buffer
        """
        if not self.__is_buffer or force:
            self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def read(self) -> List[MoleculeContainer]:
        """
        parse whole file

        :return: list of parsed molecules
        """
        return list(iter(self))

    def __iter__(self) -> Iterator[MoleculeContainer]:
        return (x for x in self._data if not isinstance(x, ParseError))

    def __next__(self) -> MoleculeContainer:
        return next(iter(self))

    def parse(self, inchi: str) -> MoleculeContainer:
        """
        Convert INCHI string into MoleculeContainer object. String should start with INCHI and
        optionally continues with space/tab separated list of key:value [or key=value] data.
        """
        if not inchi:
            raise ValueError('Empty string')

        inchi, *data = inchi.split()
        if self.__header is None:
            meta = {}
            for x in data:
                try:
                    k, v = split('[=:]', x, 1)
                    meta[k] = v
                except ValueError:
                    self._info(f'invalid metadata entry: {x}')
        else:
            meta = dict(zip(self.__header, data))

        record = self.__parse_inchi(inchi)
        record['meta'] = meta
        return self._convert_molecule(record)

    def _create_molecule(self, data, mapping):
        mol = super()._create_molecule(data, mapping, _skip_calc_implicit=True)
        atoms = mol._atoms
        bonds = mol._bonds
        charges = mol._charges
        radicals = mol._radicals
        hydrogens = mol._hydrogens
        plane = mol._plane

        # set hydrogen atoms. INCHI designed for hydrogens handling. hope correctly.
        free = count(max(mapping.values()) + 1)
        for n, atom in enumerate(data['atoms']):
            n = mapping[n]
            if atom['element'] != 'H':
                hydrogens[n] = atom['hydrogens']
            # in chython hydrogens never have implicit H.
            elif atom['hydrogens']:  # >[xH]-H case
                m = next(free)
                charges[m] = 0
                radicals[m] = False
                plane[m] = (0., 0.)
                hydrogens[n] = 0
                hydrogens[m] = 0
                atoms[m] = a = H()
                a._attach_graph(mol, m)
                bonds[n][m] = b = Bond(1)
                bonds[m] = {n: b}
                b._attach_graph(mol, n, m)
            else:  # H+, H* or >H-[xH] cases
                hydrogens[n] = 0
            # convert isotopic implicit hydrogens to explicit
            for i, k in enumerate(('p', 'd', 't'), 1):
                if atom[k]:
                    for _ in range(atom[k]):
                        m = next(free)
                        charges[m] = 0
                        radicals[m] = False
                        plane[m] = (0., 0.)
                        hydrogens[m] = 0
                        atoms[m] = a = H(i)
                        a._attach_graph(mol, m)
                        bonds[n][m] = b = Bond(1)
                        bonds[m] = {n: b}
                        b._attach_graph(mol, n, m)

        if self.__ignore_stereo or \
                not data['stereo_atoms'] and not data['stereo_cumulenes'] and not data['stereo_allenes']:
            return mol

        st = mol._stereo_tetrahedrons
        sa = mol._stereo_allenes
        ctt = mol._stereo_cis_trans_terminals

        stereo = []
        for n, ngb, s in data['stereo_atoms']:
            n = mapping[n]
            if n in st:
                stereo.append((mol.add_atom_stereo, n, [mapping[x] for x in ngb], s))
        for n, nn, mn, s in data['stereo_allenes']:
            n = mapping[n]
            if n in sa:
                stereo.append((mol.add_atom_stereo, n, mapping[nn], mapping[mn], s))
        for n, m, nn, nm, s in data['stereo_cumulenes']:
            n = mapping[n]
            if n in ctt:
                stereo.append((mol.add_cis_trans_stereo, n, mapping[m], mapping[nn], mapping[nm], s))

        while stereo:
            fail_stereo = []
            old_stereo = len(stereo)
            for f, *args in stereo:
                try:
                    f(*args, clean_cache=False)
                except NotChiral:
                    fail_stereo.append((f, *args))
                except IsChiral:
                    pass
                except ValenceError:
                    self._info('structure has errors, stereo data skipped')
                    mol.flush_cache()
                    break
            else:
                stereo = fail_stereo
                if len(stereo) == old_stereo:
                    break
                mol.flush_stereo_cache()
                continue
            break
        return mol

    @staticmethod
    def __parse_inchi(string):
        structure = INCHIStructure()
        if lib.GetStructFromINCHI(byref(InputINCHI(string)), byref(structure)):
            lib.FreeStructFromINCHI(byref(structure))
            raise ValueError('invalid INCHI')

        atoms, bonds = [], []
        seen = set()
        for n in range(structure.num_atoms):
            seen.add(n)
            atom = structure.atom[n]

            atoms.append({'element': atom.atomic_symbol, 'charge': atom.charge,
                          'mapping': 0, 'x': atom.x, 'y': atom.y, 'z': atom.z, 'isotope': atom.isotope,
                          'is_radical': atom.is_radical, 'hydrogens': atom.implicit_hydrogens,
                          'p': atom.implicit_protium, 'd': atom.implicit_deuterium, 't': atom.implicit_tritium})

            for k in range(atom.num_bonds):
                m = atom.neighbor[k]
                if m in seen:
                    continue
                order = atom.bond_type[k]
                if order:
                    bonds.append((n, m, order))

        stereo_atoms = []
        stereo_allenes = []
        stereo_cumulenes = []
        for i in range(structure.num_stereo0D):
            stereo = structure.stereo0D[i]
            sign = stereo.sign
            if sign is not None:
                if stereo.is_tetrahedral:
                    stereo_atoms.append((stereo.central_atom, stereo.neighbors, sign))
                elif stereo.is_allene:
                    nn, *_, nm = stereo.neighbors
                    stereo_allenes.append((stereo.central_atom, nn, nm, sign))
                elif stereo.is_cumulene:
                    nn, n, m, nm = stereo.neighbors
                    stereo_cumulenes.append((n, m, nn, nm, sign))

        lib.FreeStructFromINCHI(byref(structure))
        return {'atoms': atoms, 'bonds': bonds, 'stereo_atoms': stereo_atoms, 'stereo_allenes': stereo_allenes,
                'stereo_cumulenes': stereo_cumulenes}


class InputINCHI(Structure):
    def __init__(self, string, options=None):
        if options is None:
            options = b''
        else:
            options = ' '.join(f'{opt_flag}{x}' for x in options).encode()
        super().__init__(string.encode(), options)

    _fields_ = [('szInChI', c_char_p),   # InChI ASCII string to be converted to a strucure
                ('szOptions', c_char_p)  # InChI options: space-delimited; each is preceded
                                         # by '/' or '-' depending on OS and compiler
                ]


class Atom(Structure):
    @property
    def atomic_symbol(self):
        return self.elname.decode()

    @property
    def isotope(self):
        isotope = self.isotopic_mass
        if not isotope:
            isotope = None
        elif isotope > 9000:  # OVER NINE THOUSANDS!
            isotope += common_isotopes[self.atomic_symbol] - 10000
        return isotope

    @property
    def is_radical(self):
        return bool(self.radical)

    @property
    def implicit_hydrogens(self):
        return self.num_iso_H[0]

    @property
    def implicit_protium(self):
        return self.num_iso_H[1]

    @property
    def implicit_deuterium(self):
        return self.num_iso_H[2]

    @property
    def implicit_tritium(self):
        return self.num_iso_H[3]

    _fields_ = [('x', c_double), ('y', c_double), ('z', c_double),  # atom coordinates
                ('neighbor', c_short * 20),    # adjacency list: ordering numbers of the adjacent atoms, >= 0
                ('bond_type', c_byte * 20),    # inchi_BondType
                ('bond_stereo', c_byte * 20),  # inchi_BondStereo2D; negative if the sharp end points to opposite atom
                ('elname', c_char * 6),        # zero-terminated chemical element name: "H", "Si", etc.
                ('num_bonds', c_short),        # number of neighbors, bond types and bond stereo in the adjacency list
                ('num_iso_H', c_byte * 4),     # implicit hydrogen atoms
                                               # [0]: number of implicit non-isotopic H
                                               # (exception: num_iso_H[0]=-1 means INCHI adds implicit H automatically),
                                               # [1]: number of implicit isotopic 1H (protium),
                                               # [2]: number of implicit 2H (deuterium),
                                               # [3]: number of implicit 3H (tritium)
                ('isotopic_mass', c_short),    # 0 => non-isotopic; isotopic mass or 10000 + mass - average atomic mass
                ('radical', c_byte),           # inchi_Radical,
                ('charge', c_byte)]            # positive or negative; 0 => no charge


class Stereo0D(Structure):
    @property
    def is_tetrahedral(self):
        return self.type == 2

    @property
    def is_allene(self):
        return self.type == 3

    @property
    def is_cumulene(self):
        return self.type == 1

    @property
    def neighbors(self):
        return tuple(self.neighbor)

    @property
    def sign(self):
        if self.parity == 1:
            return True
        elif self.parity == 2:
            return False

    _fields_ = [('neighbor', c_short * 4),  # 4 atoms always
                ('central_atom', c_short),  # central tetrahedral atom or a central atom of allene; otherwise NO_ATOM
                ('type', c_byte),           # inchi_StereoType0D
                ('parity', c_byte)]         # inchi_StereoParity0D


class INCHIStructure(Structure):
    _fields_ = [('atom', POINTER(Atom)),          # array of num_atoms elements
                ('stereo0D', POINTER(Stereo0D)),  # array of num_stereo0D 0D stereo elements or NULL
                ('num_atoms', c_short),           # number of atoms in the structure
                ('num_stereo0D', c_short),        # number of 0D stereo elements
                ('szMessage', c_char_p),          # Error/warning ASCII message
                ('szLog', c_char_p),              # log-file ASCII string, contains a human-readable list
                                                  # of recognized options and possibly an Error/warn message
                ('WarningFlags', (c_long * 2) * 2)]

# copy-pasted from INCHI-API
#  * Notes: 1. Atom ordering numbers (i, k, and atom[i].neighbor[j] below)
#  *           start from zero; max. ordering number is (num_atoms-1).
#  *        2. inchi_Atom atom[i] is connected to the atom[atom[i].neighbor[j]]
#  *           by a bond that has type atom[i].bond_type[j] and 2D stereo type
#  *           atom[i].bond_stereo[j] (in case of no stereo
#  *           atom[i].bond_stereo[j] = INCHI_BOND_STEREO_NONE)
#  *           Index j is in the range 0 <= j <= (atom[i].num_bonds-1)
#  *        3. Any connection (represented by atom[i].neighbor[j],
#  *           atom[i].bond_type[j], and atom[i].bond_stereo[j])
#  *           should be present in one or both adjacency list:
#  *             if k = atom[i].neighbor[j] then i may or may not be present in
#  *           atom[k].neighbor[] list. For example, the adjacency lists may be
#  *           populated with only such neighbors that atom[i].neighbor[j] < i
#  *           All elements of an adjacency list must be different, that is,
#  *           a bond must be specified in an adjacency list only once.
#  *        4. in Molfiles usually
#  *           (number of implicit H) = Valence - SUM(bond_type[])
#  *        5. Seemingly illogical order of the inchi_Atom members was
#  *           chosen in an attempt to avoid alignment problems when
#  *           accessing inchi_Atom from unrelated to C programming
#  *           languages such as Visual Basic.
#  *******************************************************************/
#
# /*******************************************************************
#     0D Stereo Parity and Type definitions
#  *******************************************************************
#             Note:
#             =====
#             o Below #A is the ordering number of atom A, starting from 0
#             o See parity values corresponding to 'o', 'e', and 'u' in
#               inchi_StereoParity0D definition below)
#
#            =============================================
#             stereogenic bond >A=B< or cumulene >A=C=C=B<
#            =============================================
#
#                                  neighbor[4]  : {#X,#A,#B,#Y} in this order
#      X                           central_atom : NO_ATOM
#       \            X      Y      type         : INCHI_StereoType_DoubleBond
#        A==B         \    /
#            \         A==B
#             Y
#
#     parity= 'e'    parity= 'o'   unknown parity = 'u'
#
#     Limitations:
#     ============
#     o Atoms A and B in cumulenes MUST be connected by a chain of double bonds;
#       atoms A and B in a stereogenic 'double bond' may be connected by a double,
#       single, or alternating bond.
#     o One atom may belong to up to 3 stereogenic bonds (i.g. in a fused
#       aromatic structure).
#     o Multiple stereogenic bonds incident to any given atom should
#       either all except possibly one have (possibly different) defined
#       parities ('o' or 'e') or should all have an unknown parity 'u'.
#
#       Note on parities of alternating stereobonds
#       ===========================================
#                                                      D--E
#       In large rings  (see Fig. 1, all              //   \\
#       atoms are C) all alternating bonds         B--C      F--G
#       are treated as stereogenic.              //              \\
#       To avoid "undefined" bond parities      A                  H
#       for bonds BC, DE, FG, HI, JK, LM, AN     \               /
#       it is recommended to mark them with       N==M       J==I
#       parities.                                     \     /
#                                                       L==K    Fig. 1
#       Such a marking will make
#       the stereochemical layer unambiguous
#       and it will be different from the          B--C      F--G
#       stereochemical layer of the second       //   \\   //    \\
#       structure (Fig. 2).                     A      D--E        H
#                                                \               /
#                                                 N==M       J==I
#       By default, double and alternating            \     /
#       bonds in 8-member and greater rings             L==K    Fig. 2
#       are treated by InChI as stereogenic.
#
#
#            =============================================
#             tetrahedral atom
#            =============================================
#
#    4 neighbors
#
#             X                    neighbor[4] : {#W, #X, #Y, #Z}
#             |                    central_atom: #A
#          W--A--Y                 type        : INCHI_StereoType_Tetrahedral
#             |
#             Z
#    parity: if (X,Y,Z) are clockwize when seen from W then parity is 'e' otherwise 'o'
#    Example (see AXYZW above): if W is above the plane XYZ then parity = 'e'
#
#    3 neighbors
#
#               Y          Y       neighbor[4] : {#A, #X, #Y, #Z}
#              /          /        central_atom: #A
#          X--A  (e.g. O=S   )     type        : INCHI_StereoType_Tetrahedral
#              \          \
#               Z          Z
#
#    parity: if (X,Y,Z) are clockwize when seen from A then parity is 'e',
#                                                           otherwise 'o'
#    unknown parity = 'u'
#    Example (see AXYZ above): if A is above the plane XYZ then parity = 'e'
#    This approach may be used also in case of an implicit H attached to A.
#
#            =============================================
#             allene
#            =============================================
#
#        X       Y                 neighbor[4]  : {#X,#A,#B,#Y}
#         \     /                  central_atom : #C
#          A=C=B                   type         : INCHI_StereoType_Allene
#
#                                       Y      X
#                                       |      |
#      when seen from A along A=C=B:  X-A    Y-A
#
#                           parity:   'e'    'o'
#
#    parity: if A, B, Y are clockwise when seen from X then parity is 'e',
#                                                           otherwise 'o'
#    unknown parity = 'u'
#    Example (see XACBY above): if X on the diagram is above the plane ABY
#                                                       then parity is 'o'
#
#    Limitations
#    ===========
#    o Atoms A and B in allenes MUST be connected by a chain of double bonds;
#
#
#    How InChI uses 0D parities
#    ==========================
#
#    1. 0D parities are used if all atom coordinates are zeroes.
#
#    In addition to that:
#
#    2. 0D parities are used for Stereobonds, Allenes, or Cumulenes if:
#
#    2a. A bond to the end-atom is shorter than MIN_BOND_LEN=0.000001
#    2b. A ratio of two bond lengths to the end-atom is smaller than MIN_SINE=0.03
#    2c. In case of a linear fragment X-A=B end-atom A is treated as satisfying 2a-b
#
#        0D parities are used if 2a or 2b or 2c applies to one or both end-atoms.
#
#    3. 0D parities are used for Tetrahedral Atoms if at least one of 3a-c is true:
#
#    3a. One of bonds to the central atom is shorter than MIN_BOND_LEN=0.000001
#    3b. A ratio of two bond lengths to the central atom is smaller than MIN_SINE=0.03
#    3c. The four neighbors are almost in one plane or the central atom and
#        its only 3 explicit neighbors are almost in one plane
#
#    Notes on 0D parities and 'undefined' stereogenic elements
#    =========================================================
#
#    If 0D parity is to be used according to 1-3 but    CH3     CH3
#    has not been provided then the corresponding         \    /
#    stereogenic element is considered 'undefined'.        C=CH
#                                                         /
#    For example, if in the structure (Fig. 3)           H
#    the explicit H has been moved so that it                Fig. 3
#    has same coordinates as atom >C= (that is,
#    the length of the bond H-C became zero)
#    then the double bond is assigned 'undefined'       CH3      CH3
#    parity which by default is omitted from the          \     /
#    Identifier.                                           CH=CH
#
#    However, the structure on Fig. 4 will have double        Fig. 4
#    bond parity 'o' and its parity in the Identifier is (-).
#
#    Notes on 0D parities in structures containing metals
#    ====================================================
#    Since InChI disconnects bonds to metals the 0D parities upon the
#    disconnection may change in several different ways:
#
#    1) previously non-stereogenic bond may become stereogenic:
#
#          \     /                            \     /
#           CH==CH          disconnection      CH==CH
#            \ /               ======>
#             M                                  M
#
#      before the disconnection:    after the disconnection:
#      atoms C have valence=5 and   the double bond may become
#      the double bond is not       stereogenic
#      recognized as stereogenic
#
#    2) previously stereogenic bond may become non-stereogenic:
#
#        M                           M(+)
#         \    /                             /
#          N==C      disconnection    (-)N==C
#              \        ======>              \
#
#    3) Oddball structures, usually resulting from projecting 3D
#       structures on the plane, may contain fragment like that
#       depicted on Fig. 5:
#
#               M   A                      M   A
#               |\ /   B                      /   B
#               | X   /     disconnection    /   /
#               |/ \ /         ======>      /   /
#               C===C                      C===C
#              Fig. 5
#      (X stands for bond intersection)
#
#      A-C=C-B parity is              A-C=C-B parity is
#      trans (e)                      cis (o) or undefined
#      because the bond               because C valence = 3,
#      orientation is same            not 4.
#      as on Fig, 6 below:
#
#           A       M
#            \     /     Removal of M from the structure
#             C===C      on Fig. 5 changes the geometry from trans
#            /     \     to cis.
#           M'      B    Removal of M and M' from the structure
#           Fig. 6       on Fig. 6 does not change the A-C=C-B
#                        geometry: it is trans.
#
#    To resolve the problem InChI API accepts the second parity
#    corresponding to the metal-disconnected structure.
#    To store both bond parities use left shift by 3 bits:
#
#    inchi_Stereo0D::parity = ParityOfConnected | (ParityOfDisconnected<<3)
#
#    In case when only disconnected structure parity exists set
#    ParityOfConnected = INCHI_PARITY_UNDEFINED.
#    This is the only case when INCHI_PARITY_UNDEFINED parity
#    may be fed to the InChI.
#
#    In cases when the bond parity in a disconnected structure exists and
#    differs from the parity in the connected structure the atoms A and B
#    should be non-metals.
#


try:
    from site import getuserbase
except ImportError:
    prefixes = {prefix, exec_prefix}
else:
    user_prefix = getuserbase()
    if user_prefix:
        prefixes = {prefix, exec_prefix, user_prefix}
    else:
        prefixes = {prefix, exec_prefix}

sitepackages = []
for pr in prefixes:
    pr = Path(pr)
    if name == 'posix':
        sitepackages.append(pr / 'local' / 'lib')
    else:
        sitepackages.append(pr)
    sitepackages.append(pr / 'lib')

platform = get_platform()
if platform == 'win-amd64':
    opt_flag = '/'
    libname = 'libinchi.dll'
elif platform == 'linux-x86_64':
    opt_flag = '-'
    libname = 'libinchi.so'
elif platform.startswith('macosx') and platform.endswith('x86_64'):
    opt_flag = '-'
    libname = 'libinchi.dynlib'
else:
    warn('unsupported platform for libinchi', ImportWarning)
    libname = None
    __all__ = []
    del INCHIRead

if libname:
    for site in sitepackages:
        lib_path = site / libname
        if lib_path.exists():
            try:
                lib = cdll.LoadLibrary(str(lib_path))
            except OSError:
                warn('libinchi loading problem', ImportWarning)
                __all__ = []
                del INCHIRead
                break
            __all__ = ['INCHIRead']
            break
    else:
        warn('broken package installation. libinchi not found', ImportWarning)
        __all__ = []
        del INCHIRead



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Dinar Batyrshin <batyrshin-dinar@mail.ru>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from bisect import bisect_left
from collections import defaultdict
from itertools import chain
from os.path import getsize
from subprocess import check_output
from time import strftime
from traceback import format_exc
from typing import Union
from ._mdl import MDLRead, MDLWrite, MOLRead, EMOLRead, RXNRead, ERXNRead, EMDLWrite
from ..containers import ReactionContainer, MoleculeContainer
from ..exceptions import ParseError, ParseReactionError


class FallBack:
    def __init__(self, is_reaction, meta=None):
        self.is_reaction = is_reaction
        self.__meta = {} if meta is None else meta

    def __call__(self, line):
        return line.startswith(('M  END', 'M  V30 END CTAB'))

    def getvalue(self):
        if self.is_reaction:
            # ad-hocs for raising ValueError
            return {'reactants': None, 'products': None, 'reagents': None, 'meta': {}}
        return {'meta': self.__meta}


class RDFRead(MDLRead):
    """
    MDL RDF files reader. works similar to opened file object. support `with` context manager.
    on initialization accept opened in text mode file, string path to file,
    pathlib.Path object or another buffered reader object
    """
    def __init__(self, file, indexable=False, **kwargs):
        """
        :param indexable: if True: supported methods seek, tell, object size and subscription, it only works when
            dealing with a real file (the path to the file is specified) because the external grep utility is used,
            supporting in unix-like OS the object behaves like a normal open file.

            if False: works like generator converting a record into ReactionContainer and returning each object in
            order, records with errors are skipped
        :param ignore: Skip some checks of data or try to fix some errors.
        :param remap: Remap atom numbers started from one.
        :param store_log: Store parser log if exists messages to `.meta` by key `ParserLog`.
        :param calc_cis_trans: Calculate cis/trans marks from 2d coordinates.
        :param ignore_stereo: Ignore stereo data.
        :param ignore_bad_isotopes: reset invalid isotope mark to non-isotopic.
        """
        super().__init__(file, **kwargs)
        self.__file = iter(self._file.readline, '')
        self._data = self.__reader()

        if indexable:
            if next(self._data):
                self._load_cache()
        else:
            next(self._data)

    @staticmethod
    def _get_shifts(file):
        shifts = [int(x.split(b':', 1)[0]) for x in
                  check_output(['grep', '-boE', r'^\$[RM]FMT', file]).split()]
        shifts.append(getsize(file))
        return shifts

    def seek(self, offset):
        """
        shifts on a given number of record in the original file
        :param offset: number of record
        """
        if self._shifts:
            if 0 <= offset < len(self._shifts):
                current_pos = self._file.tell()
                new_pos = self._shifts[offset]
                if current_pos != new_pos:
                    if current_pos == self._shifts[-1]:  # reached the end of the file
                        self._file.seek(0)
                        self.__file = iter(self._file.readline, '')
                        self._data = self.__reader()
                        next(self._data)
                        if offset:  # move not to the beginning of the file
                            self._file.seek(new_pos)
                            self._data.send(offset)
                            self.__already_seeked = True
                    else:
                        if not self.__already_seeked:
                            self._file.seek(new_pos)
                            self._data.send(offset)
                            self.__already_seeked = True
                        else:
                            raise BlockingIOError('File already seeked. New seek possible only after reading any data')
            else:
                raise IndexError('invalid offset')
        else:
            raise self._implement_error

    def tell(self):
        """
        :return: number of records processed from the original file
        """
        if self._shifts:
            t = self._file.tell()
            if t == self._shifts[0]:
                return 0
            elif t == self._shifts[-1]:
                return len(self._shifts) - 1
            elif t in self._shifts:
                return bisect_left(self._shifts, t)
            else:
                return bisect_left(self._shifts, t) - 1
        raise self._implement_error

    def __reader(self):
        record = parser = mkey = pos = None
        failed = False
        file = self._file
        try:
            seekable = file.seekable()
        except AttributeError:
            seekable = False

        if next(self.__file).startswith('$RXN'):  # parse RXN file
            is_reaction = True
            ir = 3
            meta = defaultdict(list)
            if seekable:
                pos = 0  # $RXN line starting position
            count = 0
            yield False
        elif next(self.__file).startswith('$DATM'):  # skip header
            ir = 0
            is_reaction = meta = None
            seek = yield True
            if seek is not None:
                yield
                count = seek - 1
                self.__already_seeked = False
            else:
                count = -1
        else:
            raise ValueError('invalid file')

        for line in self.__file:
            if failed and not line.startswith(('$RFMT', '$MFMT')):
                continue
            elif parser:
                try:
                    if parser(line):
                        record = parser.getvalue()
                        parser = None
                except ValueError:
                    self._info(f'line:\n{line}\nconsist errors:\n{format_exc()}')
                    if self._store_log:
                        if isinstance(parser, EMOLRead):  # try to restore metadata
                            parser = FallBack(False, parser._meta)  # noqa
                        else:
                            parser = FallBack(is_reaction)
                        continue
                    parser = None
                    seek = yield ParseError(count, pos, self._format_log(), None)
                    if seek is not None:
                        yield
                        count = seek - 1
                        self.__already_seeked = False
                    else:
                        failed = True
            elif line.startswith('$RFMT'):
                if record:
                    record['meta'].update(self._prepare_meta(meta))
                    if title:
                        record['title'] = title
                    try:
                        if is_reaction:
                            container = self._convert_reaction(record)
                        elif 'atoms' not in record:  # mol fall-back ad-hoc
                            record['meta']['chytorch_mdl_title'] = title
                            container = ParseError(count, pos, self._format_log(), record['meta'])
                        else:
                            container = self._convert_molecule(record)
                    except ParseReactionError as e:  # ad-hoc for logging partially parsed reactions
                        record['meta']['chytorch_mdl_title'] = title
                        seek = yield ParseError(count, pos, self._format_log(), record['meta'], e.structure, e.errors)
                    except ValueError:
                        self._info(f'record consist errors:\n{format_exc()}')
                        record['meta']['chytorch_mdl_title'] = title
                        seek = yield ParseError(count, pos, self._format_log(), record['meta'])
                    else:
                        seek = yield container

                    record = None
                    if seek is not None:
                        yield
                        count = seek - 1
                        self.__already_seeked = False
                        continue

                if seekable:
                    pos = file.tell()  # $RXN line starting position
                count += 1
                is_reaction = True
                ir = 4
                failed = False
                mkey = None
                meta = defaultdict(list)
            elif line.startswith('$MFMT'):
                if record:
                    record['meta'].update(self._prepare_meta(meta))
                    if title:
                        record['title'] = title
                    try:
                        if is_reaction:
                            container = self._convert_reaction(record)
                        elif 'atoms' not in record:  # mol fall-back ad-hoc
                            record['meta']['chytorch_mdl_title'] = title
                            container = ParseError(count, pos, self._format_log(), record['meta'])
                        else:
                            container = self._convert_molecule(record)
                    except ParseReactionError as e:  # ad-hoc for logging partially parsed reactions
                        record['meta']['chytorch_mdl_title'] = title
                        seek = yield ParseError(count, pos, self._format_log(), record['meta'], e.structure, e.errors)
                    except ValueError:
                        self._info(f'record consist errors:\n{format_exc()}')
                        record['meta']['chytorch_mdl_title'] = title
                        seek = yield ParseError(count, pos, self._format_log(), record['meta'])
                    else:
                        seek = yield container

                    record = None
                    if seek is not None:
                        yield
                        count = seek - 1
                        self.__already_seeked = False
                        continue

                if seekable:
                    pos = file.tell()  # MOL block line starting position
                count += 1
                ir = 3
                failed = is_reaction = False
                mkey = None
                meta = defaultdict(list)
            elif record:
                if line.startswith('$DTYPE'):
                    mkey = line[7:].strip()
                    if not mkey:
                        self._info(f'invalid metadata entry: {line}')
                elif mkey:
                    data = line.lstrip("$DATUM").strip()
                    if data:
                        meta[mkey].append(data)
            elif ir:
                if ir == 3:  # parse mol or rxn title
                    title = line.strip()
                ir -= 1
            else:
                try:
                    if is_reaction:
                        if line.startswith('M  V30 COUNTS'):
                            parser = ERXNRead(line, self._ignore, self._log_buffer)
                        else:
                            parser = RXNRead(line, self._ignore, self._log_buffer)
                    else:
                        if 'V2000' in line:
                            parser = MOLRead(line, self._log_buffer)
                        elif 'V3000' in line:
                            parser = EMOLRead(self._ignore, self._log_buffer)
                        else:
                            raise ValueError('invalid MOL entry')
                except ValueError:
                    self._info(f'line:\n{line}\nconsist errors:\n{format_exc()}')
                    if self._store_log:
                        parser = FallBack(is_reaction)
                        continue
                    failed = True
                    seek = yield ParseError(count, pos, self._format_log(), None)
                    if seek is not None:
                        yield
                        count = seek - 1
                        self.__already_seeked = False
        if record:
            record['meta'].update(self._prepare_meta(meta))
            if title:
                record['title'] = title
            try:
                if is_reaction:
                    container = self._convert_reaction(record)
                elif 'atoms' not in record:  # mol fall-back ad-hoc
                    record['meta']['chytorch_mdl_title'] = title
                    container = ParseError(count, pos, self._format_log(), record['meta'])
                else:
                    container = self._convert_molecule(record)
            except ParseReactionError as e:  # ad-hoc for logging partially parsed reactions
                record['meta']['chytorch_mdl_title'] = title
                yield ParseError(count, pos, self._format_log(), record['meta'], e.structure, e.errors)
            except ValueError:
                self._info(f'record consist errors:\n{format_exc()}')
                record['meta']['chytorch_mdl_title'] = title
                yield ParseError(count, pos, self._format_log(), record['meta'])
            else:
                yield container

    __already_seeked = False


class _RDFWrite:
    def __init__(self, file, *, append: bool = False, mapping: bool = True):
        """
        :param append: append to existing file (True) or rewrite it (False). For buffered writer object append = False
            will write RDF header and append = True will omit the header.
        :param mapping: write atom mapping.
        """
        super().__init__(file, append=append, mapping=mapping)
        if not append or not (self._is_buffer or self._file.tell() != 0):
            self.write = self.__write

    def __write(self, data):
        """
        write single molecule or reaction into file
        """
        del self.write
        self._file.write(strftime('$RDFILE 1\n$DATM    %m/%d/%y %H:%M\n'))
        self.write(data)


class RDFWrite(_RDFWrite, MDLWrite):
    """
    MDL RDF files writer. works similar to opened for writing file object. support `with` context manager.
    on initialization accept opened for writing in text mode file, string path to file,
    pathlib.Path object or another buffered writer object
    """
    def write(self, data: Union[ReactionContainer, MoleculeContainer]):
        if isinstance(data, ReactionContainer):
            ag = f'{len(data.reagents):3d}' if data.reagents else ''
            self._file.write(f'$RFMT\n$RXN\n{data.name}\n\n\n{len(data.reactants):3d}{len(data.products):3d}{ag}\n')
            for m in chain(data.reactants, data.products, data.reagents):
                m = self._convert_molecule(m)
                self._file.write('$MOL\n')
                self._file.write(m)
        else:
            m = self._convert_molecule(data)
            self._file.write('$MFMT\n')
            self._file.write(m)
        for k, v in data.meta.items():
            self._file.write(f'$DTYPE {k}\n$DATUM {v}\n')


class ERDFWrite(_RDFWrite, EMDLWrite):
    """
    MDL V3000 RDF files writer. works similar to opened for writing file object. support `with` context manager.
    on initialization accept opened for writing in text mode file, string path to file,
    pathlib.Path object or another buffered writer object
    """
    def write(self, data: Union[ReactionContainer, MoleculeContainer]):
        if isinstance(data, ReactionContainer):
            ag = f' {len(data.reagents)}' if data.reagents else ''
            self._file.write(f'$RFMT\n$RXN V3000\n{data.name}\n\n\n'
                             f'M  V30 COUNTS {len(data.reactants)} {len(data.products)}{ag}\nM  V30 BEGIN REACTANT\n')
            for m in data.reactants:
                self._file.write(self._convert_molecule(m))
            self._file.write('M  V30 END REACTANT\nM  V30 BEGIN PRODUCT\n')
            for m in data.products:
                self._file.write(self._convert_molecule(m))
            self._file.write('M  V30 END PRODUCT\n')
            if data.reagents:
                self._file.write('M  V30 BEGIN AGENT\n')
                for m in data.reagents:
                    self._file.write(self._convert_molecule(m))
                self._file.write('M  V30 END AGENT\n')
            self._file.write('M  END\n')
        else:
            m = self._convert_molecule(data)
            self._file.write(f'$MFMT\n{data.name}\n\n\n  0  0  0     0  0            999 V3000\n')
            self._file.write(m)
            self._file.write('M  END\n')
        for k, v in data.meta.items():
            self._file.write(f'$DTYPE {k}\n$DATUM {v}\n')


__all__ = ['RDFRead', 'RDFWrite', 'ERDFWrite']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABCMeta
from .element import *
from .groups import *
from .periods import *
from .groupI import *
from .groupII import *
from .groupIII import *
from .groupIV import *
from .groupV import *
from .groupVI import *
from .groupVII import *
from .groupVIII import *
from .groupIX import *
from .groupX import *
from .groupXI import *
from .groupXII import *
from .groupXIII import *
from .groupXIV import *
from .groupXV import *
from .groupXVI import *
from .groupXVII import *
from .groupXVIII import *

modules = {v.__name__: v for k, v in globals().items() if k.startswith('group') and k != 'groups'}
elements = {k: v for k, v in globals().items() if isinstance(v, ABCMeta) and k != 'Element' and issubclass(v, Element)}

__all__ = ['Element', 'DynamicElement', 'QueryElement', 'AnyElement', 'ListElement', 'AnyMetal']
__all__.extend(k for k in globals() if k.startswith('Group'))
__all__.extend(k for k in globals() if k.startswith('Period'))
__all__.extend(elements)


for _class in (DynamicElement, QueryElement):
    for k, v in elements.items():
        name = f'{_class.__name__[:-7]}{k}'
        globals()[name] = cls = type(name, (_class, *v.__mro__[-3:-1]),
                                     {'__module__': v.__module__, '__slots__': (), 'atomic_number': v.atomic_number,
                                      'atomic_radius': v.atomic_radius})
        setattr(modules[v.__module__], name, cls)
        modules[v.__module__].__all__.append(name)
        __all__.append(name)



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.math cimport ldexp, frexp

# Format specification::
#
# Big endian bytes order
# 8 bit - 0x02 (current format specification)
# 12 bit - number of atoms
# 12 bit - cis/trans stereo block size
#
# Atom block 9 bytes (repeated):
# 12 bit - atom number
# 4 bit - number of neighbors
# 2 bit tetrahedron sign (00 - not stereo, 10 or 11 - has stereo)
# 2 bit - allene sign
# 5 bit - isotope (00000 - not specified, over = isotope - common_isotope + 16)
# 7 bit - atomic number (<=118)
# 32 bit - XY float16 coordinates
# 3 bit - hydrogens (0-7). Note: 7 == None
# 4 bit - charge (charge + 4. possible range -4 - 4)
# 1 bit - radical state
# Connection table: flatten list of neighbors. neighbors count stored in atom block.
# For example CC(=O)O - {1: [2], 2: [1, 3, 4], 3: [2], 4: [2]} >> [2, 1, 3, 4, 2, 2].
# Repeated block (equal to bonds count).
# 24 bit - paired 12 bit numbers.
# Bonds order block 3 bit per bond zero-padded to full byte at the end.
# Cis/trans data block (repeated):
# 24 bit - atoms pair
# 7 bit - zero padding. in future can be used for extra bond-level stereo, like atropoisomers.
# 1 bit - sign

@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def pack(object molecule):
    cdef bint b  # binary flag
    cdef char charge
    cdef unsigned char atomic_number, isotope, bond, s = 0, buffer_b, buffer_o
    cdef unsigned char *p, *data
    cdef unsigned short atoms_count, bonds_count = 0, cis_trans_count, n, m
    cdef unsigned int size, atoms_shift = 4, bonds_shift, order_shift, cis_trans_shift  # can be > 2^16
    cdef unsigned char[4096] stereo, hcr, seen
    cdef unsigned int[4096] xy  # 2 * 16bit

    cdef bytes py_pack
    cdef dict py_ngb, py_atoms, py_bonds, py_charges, py_radicals, py_hydrogens, py_plane
    cdef dict py_cis_trans_stereo, py_atoms_stereo, py_allenes_stereo
    cdef tuple py_tuple
    cdef object py_atom, py_bond, py_nan_int, py_obj

    # map molecule to vars
    py_atoms = molecule._atoms
    py_bonds = molecule._bonds
    py_charges = molecule._charges
    py_radicals = molecule._radicals
    py_hydrogens = molecule._hydrogens
    py_cis_trans_stereo = molecule._cis_trans_stereo
    py_atoms_stereo = molecule._atoms_stereo
    py_allenes_stereo = molecule._allenes_stereo
    py_plane = molecule._plane

    # calculate elements count
    atoms_count = len(py_atoms)
    cis_trans_count = len(py_cis_trans_stereo)

    for py_ngb in py_bonds.values():
        bonds_count += len(py_ngb)
    bonds_count /= 2  # graph is bidirected

    # calculate pack blocks entries
    size = bonds_count * 3  # bonds bits
    if size % 8:  # partial byte fill
        size = size / 8 + 1
    else:
        size /= 8
    bonds_shift = 4 + 9 * atoms_count  # connection table starting byte
    order_shift = bonds_shift + 3 * bonds_count  # bond orders block starting byte
    cis_trans_shift = size + order_shift  # cis-trans block starting byte
    size = cis_trans_shift + 4 * cis_trans_count  # total pack size

    # allocate pack in memory
    data = <unsigned char *> PyMem_Malloc(size * sizeof(unsigned char))
    if not data:
        raise MemoryError()

    # precalculate atom attrs
    # should be done independently, due to possible randomness in dicts order.
    # 3 bit - hydrogens (0-7) | 4 bit - charge | 1 bit - radical
    for n, py_nan_int in py_hydrogens.items():
        if py_nan_int is None:
            hcr[n] = 0xe0  # 0b11100000
        else:
            hcr[n] = <unsigned char> py_nan_int << 5
    for n, charge in py_charges.items():
        hcr[n] |= (charge + 4) << 1
    for n, b in py_radicals.items():
        if b:  # lazy memory access
            hcr[n] |= 1

    # 2 float16 big endian
    for n, py_tuple in py_plane.items():
        p = <unsigned char *> &xy[n]
        double_to_float16(py_tuple[0], &p[0])
        double_to_float16(py_tuple[1], &p[2])

        # erase random data
        seen[n] = 0
        stereo[n] = 0

    # 2 bit tetrahedron | 2 bit allene | 0000
    for n, b in py_atoms_stereo.items():
        stereo[n] = 0xc0 if b else 0x80
    for n, b in py_allenes_stereo.items():
        stereo[n] = 0x30 if b else 0x20

    # start pack collection
    data[0] = 2  # header. specification version 2
    data[1] = atoms_count >> 4  # 5-12b of atom count value
    data[2] = atoms_count << 4 | cis_trans_count >> 8  # 1-4b of atom count value, 9-12b of cis-trans count value
    data[3] = cis_trans_count  # 1-8b of cis-trans count value

    b = True  # init connection table flag
    for py_obj, py_atom in py_atoms.items():
        py_ngb = py_bonds[py_obj]
        n = py_obj  # cast to C
        seen[n] = 1
        p = <unsigned char *> &xy[n]  # XY
        atomic_number = py_atom.atomic_number
        py_nan_int = py_atom._Core__isotope  # direct access
        if py_nan_int is None:
            isotope = 0
        else:
            isotope = <short> py_nan_int - common_isotopes[atomic_number]

        data[atoms_shift] = n >> 4  # 5-12b AN
        data[atoms_shift + 1] = n << 4 | len(py_ngb)  # 1-4b AN, 4b NC
        data[atoms_shift + 2] = stereo[n] | isotope >> 1  # TS , AS , 4b I
        data[atoms_shift + 3] = isotope << 7 | atomic_number  # 1bI , A
        data[atoms_shift + 4] = p[0]
        data[atoms_shift + 5] = p[1]
        data[atoms_shift + 6] = p[2]
        data[atoms_shift + 7] = p[3]
        data[atoms_shift + 8] = hcr[n]
        atoms_shift += 9

        # collect connection table
        for m, py_bond in py_ngb.items():
            if b:  # 8 + 4
                data[bonds_shift] = m >> 4
                bonds_shift += 1
                buffer_b = m << 4
                b = False  # switch
            else:  # 4 + 8
                data[bonds_shift] = buffer_b | m >> 8
                bonds_shift += 1
                data[bonds_shift] = m
                bonds_shift += 1  # next free 3 bytes block
                b = True

            if not seen[m]:
                bond = <unsigned char> py_bond._Bond__order - 1
                # 3 3 2 | 1 3 3 1 | 2 3 3
                if s == 0:
                    buffer_o = bond << 5
                    s = 1
                elif s == 1:
                    buffer_o |= bond << 2
                    s = 2
                elif s == 2:
                    data[order_shift] = buffer_o | bond >> 1
                    order_shift += 1
                    buffer_o = bond << 7
                    s = 3
                elif s == 3:
                    buffer_o |= bond << 4
                    s = 4
                elif s == 4:
                    buffer_o |= bond << 1
                    s = 5
                elif s == 5:
                    data[order_shift] = buffer_o | bond >> 2
                    order_shift += 1
                    buffer_o = bond << 6
                    s = 6
                elif s == 6:
                    buffer_o |= bond << 3
                    s = 7
                else:  # 7
                    data[order_shift] = buffer_o | bond
                    order_shift += 1
                    s = 0

    if s:  # flush buffer
        data[order_shift] = buffer_o

    for py_tuple, b in py_cis_trans_stereo.items():
        n, m = py_tuple
        data[cis_trans_shift] = n >> 4
        data[cis_trans_shift + 1] = n << 4 | m >> 8
        data[cis_trans_shift + 2] = m
        data[cis_trans_shift + 3] = b
        cis_trans_shift += 4

    try:
        py_pack = data[:size]
    finally:
        PyMem_Free(data)
    return py_pack


cdef short[119] common_isotopes
common_isotopes[:] = [0, -15, -12, -9, -7, -5, -4, -2, 0, 3, 4, 7, 8, 11, 12, 15, 16, 19, 24, 23, 24, 29,
                      32, 35, 36, 39, 40, 43, 43, 48, 49, 54, 57, 59, 63, 64, 68, 69, 72, 73, 75, 77,
                      80, 82, 85, 87, 90, 92, 96, 99, 103, 106, 112, 111, 115, 117, 121, 123, 124, 125,
                      128, 129, 134, 136, 141, 143, 147, 149, 151, 153, 157, 159, 162, 165, 168, 170,
                      174, 176, 179, 181, 185, 188, 191, 193, 193, 194, 206, 207, 210, 211, 216, 215,
                      222, 221, 228, 227, 231, 231, 235, 236, 241, 242, 243, 244, 245, 254, 253, 254,
                      254, 262, 265, 265, 269, 262, 273, 273, 277, 281, 278]


cdef void double_to_float16(double x, unsigned char* p):
    # adopted from cpython source code
    cdef unsigned char sign
    cdef int e
    cdef double f
    cdef unsigned short bits

    if x == 0.:
        p[0] = p[1] = 0
        return

    sign = x < 0.
    if sign:
        x = -x
    f = frexp(x, &e)
    e -= 1
    if f < .5 or f >= 1. or e >= 16 or e < -25:
        p[0] = p[1] = 0
        return  # ignore big values

    f *= 2.0
    if e < -14:
        f = ldexp(f, 14 + e)
        e = 0
    else:
        e += 15
        f -= 1.

    f *= 1024.
    bits = <unsigned short> f | (e << 10) | (sign << 15)
    p[0] = bits >> 8
    p[1] = bits
